#!/usr/bin/python

####
# Re-implementation of the original SALT2 model.
####

import time
import numpy as np
import scipy.sparse

from ..lib.fitparameters import FitParameters
from ..lib import bspline
from ..instruments import FilterDb
import logging
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)


class ColorLaw(object):
    r"""Implementation of the SALT2 color law (version 1) 

    The SALT2 color law describes the color diversity of SNeIa. In
    SALT 2.4, the color law is parametrized as follows:
    
    .. math::
        CL(\lambda;c) = 10^{0.4\ c \times P(\lambda_r)}

    where :math:`P(\lambda_r)` is a polynomial, such as
    :math:`P(\lambda_B) = 0` and :math:`P(\lambda_V) = 1`.

    This implies the following transformation :

    .. math::
        \lambda_r = \frac{\lambda-\lambda^B}{\lambda^V - \lambda^B}

    It is defined in the wavelength range :math:`2800 A < \lambda < 7000 A`
    and extrapolated with a linear function outside this interval.

    This class evaluates :math:`P(\lambda, \alpha_i)` and its
    derivatives :math:`\partial P/\partial \alpha_i` (which is the
    painful part).

    Attributes
    ----------
    min_lambda : float
        Lower value of the wavelength interval
    max_lambda : float
        Upper value of the wavelength interval
    min_reduced_lambda : float
        Lower value of the reduce wavelength interval
    max_reduced_lambda : float
        Upper value of the reduce wavelength interval
    pars : array
        Color law polynomial parameters
    pars_ir :
        Parameters of the IR affine extension
    pars_uv
        Parameters of the UV affine extension
    """
    WAVELENGTH = {"U": 3650.88, "B": 4302.57, "V": 5428.55, "R": 6418.01, "I": 7968.34}
    U_WAVELENGTH = 3650.88
    B_WAVELENGTH = 4302.57
    V_WAVELENGTH = 5428.55
    R_WAVELENGTH = 6418.01
    I_WAVELENGTH = 7968.34

    def __init__(self, wl_range=[2800, 7000]):
        """
        Constructor - computes the arguments

        Parameters
        ----------
        wl_range : list or ndarray
            nominal wavelength range for the polynom color law.
            Outside this range, we don't expect to have valid data
            and the color law is extrapolated with deg-1 polynomials.
        """
        assert(wl_range[0] < wl_range[1])
        self.min_lambda, self.max_lambda = wl_range
        
        self.min_reduced_lambda = self.reduce(self.min_lambda)
        self.max_reduced_lambda = self.reduce(self.max_lambda)
        self.pars = np.zeros(5)

    def reduce(self, wl):
        r"""Linear remapping of the wavelength passed in argument. 
        
        The remapping is defined by:

        .. math::
            \lambda_r = \frac{\lambda - \lambda_B}{\lambda_V - \lambda_B}

        Parameters
        ----------
        wl : array-like
            the input wavelengths in Angstroms

        Returns
        ----------
        array of floats
            the reduced wavelengths
        """
        B_WL, V_WL = self.WAVELENGTH["B"], self.WAVELENGTH["V"]
        return (wl-B_WL)/(V_WL-B_WL)
    
    def __call__(self, wl, p, jac=False):  #  , return_jacobian_as_coo_matrix=False):
        r"""Evaluate the polynomial part of the color law 
        
        The full SALT2 color law is given by:

        .. math::
            CL(\lambda, c) = 10^{0.4\ c\ P(\lambda, \alpha_i)}

        with 

        .. math::
            P(\lambda) = \left\{ \begin{split}
            P'(\bar\lambda_{UV}) \times (\bar\lambda-\bar\lambda_{UV}) + P(\bar\lambda_{UV})
            & \ \ \ \ \mathrm{if \lambda < \lambda_{UV}} \\
            \bar\lambda \times \left(\sum_{i=1}^4 \alpha_i \bar\lambda^i + 1 - \sum_{i=1}^4\alpha_i\right)
            & \ \ \ \ \mathrm{if \lambda_{UV} < \lambda < \lambda_{IR}} \\
            P'(\bar\lambda_{IR}) \times (\bar\lambda-\bar\lambda_{IR}) + P(\bar\lambda_{IR})
            & \ \ \ \ \mathrm{if \lambda > \lambda_{IR}} \\
            \end{split}\right.

        and :math:`\lambda_r` defined above.
        
        This function evaluates :math:`P(\lambda,\alpha_i)` along with its derivatives w.r.t. the :math:`\alpha_i`:
        :math:`\partial P/\partial \alpha_i`
        
        Parameters
        ----------
        wl : ndarray of float
            input wavelengths (in Angstroms)
        p : ndarray
            color law parameters (i.e. the :math:`(\alpha_i)_{1\leq i \leq 4}`)
        jac : bool (default=False)
            whether to return the jacobian matrix
        return_jacobian_as_coo_matrix : bool 
            If jacobian needs to be scipy.sparse

        Returns
        ----------
        cl : array-like of float
            the color law values
        jacobian : numpy.array or scipy.sparse.csr_matrix (return_jacobian_as_coo_matrix)
            the matrix of derivatives, if `jac` is True
        """
        self.pars[0:4] = p
        self.pars[4] = 1 - p.sum()
        d_pars = np.polyder(self.pars)

        # nominal range
        rwl = self.reduce(np.asarray(wl))
        r = np.polyval(self.pars, rwl) * rwl

        # uv side (if necessary)
        idx_uv = rwl < self.min_reduced_lambda
        has_uv_data = idx_uv.sum() > 0
        if has_uv_data:
            val = np.polyval(self.pars, self.min_reduced_lambda)
            d_val = np.polyval(d_pars, self.min_reduced_lambda)
            self.pars_uv = np.asarray([d_val * self.min_reduced_lambda + val,
                                       val * self.min_reduced_lambda])
            r[idx_uv] = np.polyval(self.pars_uv, rwl[idx_uv]-self.min_reduced_lambda)
        
        # ir side
        idx_ir = rwl > self.max_reduced_lambda
        has_ir_data = idx_ir.sum() > 0
        if has_ir_data:
            val = np.polyval(self.pars, self.max_reduced_lambda)
            d_val = np.polyval(d_pars, self.max_reduced_lambda)
            self.pars_ir = np.asarray([d_val * self.max_reduced_lambda + val,
                                       val * self.max_reduced_lambda])
            r[idx_ir] = np.polyval(self.pars_ir, rwl[idx_ir]-self.max_reduced_lambda)

        if not jac:
            return r, None

        # the jacobian is unfortunately a dense matrix
        # (maybe we should try using splines)
        #
        # In the nominal wavelength range, it has the form:
        # 
        #   [l_1**5-l_1   l_1**4-l1    l_1**3-l1  l_1**2-l1]
        #   [  ...          ...           ...       ...    ]
        #   [l_N**5-l_1   l_N**4-l1    l_N**3-l1  l_N**2-l1]
        # 
        v = np.vander(rwl, 6)[:, 0:-2]
        jacobian = (v.T-rwl).T

        # 
        # and in the extrapolation range, it has the form:
        #
        # J_ik = [dCL'/da_k(rwl-rwl_uv) + dCL/da_k]
        #
        # Granted, it's not very readable. But it should be fast enough 
        if has_uv_data:
            l_uv = self.min_reduced_lambda
            j1_luv = (np.vander([l_uv], 6)[:, 0:-2].T - np.array([l_uv])).T
            j2_luv = np.vander([l_uv], 5)[:, 0:-1] * np.array([5., 4., 3., 2.]) - 1.
            n_uv = idx_uv.sum()
            jacobian[idx_uv] = j2_luv * (rwl[idx_uv]-l_uv).reshape(n_uv, -1) + j1_luv * np.ones(n_uv).reshape(n_uv, -1)
            
        if has_ir_data:
            l_ir = self.max_reduced_lambda
            j1_lir = (np.vander([l_ir], 6)[:, 0:-2].T - np.array([l_ir])).T
            j2_lir = np.vander([l_ir], 5)[:, 0:-1] * np.array([5., 4., 3., 2.]) - 1.
            n_ir = idx_ir.sum()
            jacobian[idx_ir] = j2_lir * (rwl[idx_ir]-l_ir).reshape(n_ir, -1) + j1_lir * np.ones(n_ir).reshape(n_ir, -1)

        # check for nan's and inf's
        if np.any(np.isnan(r)) or np.any(np.isinf(r)):
            raise ValueError('inf/nan values in color law')

        #if return_jacobian_as_coo_matrix:
        #    jacobian = scipy.sparse.coo_matrix(jacobian)
        
        return r, jacobian
    

class SpectrumRecalibrationPolynomials:
    r"""A class to manage the spectrum recalibration polynomials
    The photometric calibration of spectra is generally affected by significant
    error modes at large wavelength scales.
    It is imperative to remove these error modes, while preserving the information provided by
    the spectra at small scales (spectral features)

    This is achieved during training by multiplying the spectral predictions of the model by a
    recalibration polynomial specific to each spectrum, function of the observation
    wavelength :math:`\lambda_o`, and of order :math:`N_s = 3`, common to all spectra:

    .. math::

        s(\lambda_{rec}) = \sum_i^{N_s} s_i \lambda_{rec}^{N_s - i} %quad \mbox{and}
        \quad \lambda_{0} = \frac{\lambda - 5000}{9000 - 2700}

    with:

    .. math::
        \lambda_{rec} = \frac{2 (\lambda_o - \lambda_{max})}{\lambda_{max} - \lambda_{min}+1

    Attributes
    ----------
    tds : nacl.dataset.TrainingDataset
        Data set of photometric and spectroscopic observations.
    model : nacl.model.salt
        Model.
    pol_degrees : int or indexable structure
        Polynomial degree for each spectrum.
    N : int
        Number of spectral data point.
    offset : numpy.array
        Starting point of parameter index for each spectrum.
    n : int
        Number of recalibration parameters
    jacobian : scipy.sparse.coo_matrix
        Jacobian matrix of the recalibration polynomial functions.
    """
    def __init__(self, tds, model, pol_degrees):
        """Constructor 

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.model.salt
            Model.
        pol_degrees : int or indexable structure
            Polynomial degree for each spectrum.
        """
        self.tds = tds
        self.model = model
        self.pol_degrees = pol_degrees
        self.jacobian = None
        
        self.N = len(self.tds.sp_data)
        o = np.cumsum(np.hstack(([0], self.pol_degrees+1)))
        self.offset, self.n = o[:-1], o[-1]
        self.build_jacobian_matrix()
        

    def init_pars(self):
        """
        return a parameter vector initialized such that
        the recalibration polynomials are evaluated to 1
        for each spectra.

        Returns
        -------
        array
            Initiated parameters.
        """
        p = np.zeros(self.n)
        o = np.cumsum(self.pol_degrees+1)
        p[o] = 1.
        return p
        
    def build_jacobian_matrix(self):
        """
        Create the jacobian matrix where line correspond to all spectral data point and
        the colones to parameters.
        """
        i, j, v = [], [], []

        for isp in range(self.tds.nb_spectra()):
            # for sp in self.tds.spectra:
            idx_sp = self.tds.spec_data['spec_id'] == isp
            sp = self.tds.spec_data[idx_sp]
            sli = np.where(idx_sp)[0]
            slc = slice(sli.min(), sli.max()+1)

            lmin, lmax = sp['Wavelength'].min(), sp['Wavelength'].max()
            a = 2. / (lmax-lmin)
            b = 1 - 2. * lmax / (lmax-lmin)
            rwl = a * sp['Wavelength'] + b
            deg = self.pol_degrees[isp]
            jacobian_spec_rec = scipy.sparse.coo_matrix(np.vander(rwl, deg+1))
            i.append(jacobian_spec_rec.row + slc.start)
            j.append(jacobian_spec_rec.col + self.offset[isp])
            v.append(jacobian_spec_rec.data)

        i = np.hstack(i)
        j = np.hstack(j)
        v = np.hstack(v)
        self.jacobian = scipy.sparse.coo_matrix((v, (i, j)), shape=(self.N, self.n)).tocsr()
                             
    def __call__(self, jac=False):
        """Evaluate the recalibration polynomials

        Parameters
        ----------
        jac : bool
            If the jacobian is needed.

        Returns
        -------
        v : numpy.array
            Recalibration parameter evaluated.
        self.jacobian : None or scipy.sparse.coo_matrix
            Jacobian matrix of the recalibration polynomial functions.
        """
        p_full = self.model.pars['SpectrumRecalibration'].full
        v = self.jacobian.dot(p_full)
        if not jac:
            return v, None
        return v, self.jacobian

    
class SALT2Like(object):
    r"""A re-implementation of the SALT2 model 
    
    SALT2 is an empirical SN spectrophotometric model. This class
    provides a pure python re-implementation of SALT2 with a number of
    improvements.

    The SALT2 parametrization is defined as follows. In the SN restframe
    the absolute spectral density :math:`S(\lambda)` is:

    .. math::   

          S(\lambda, \mathrm{p}) = X_0 \times \left[M_0(\lambda, \mathrm{p}) + X_1
          \times M_1(\lambda, \mathrm{p})\right]\ 10^{0.4 c CL(\lambda)}
    
    where :math:`\mathrm{p}` is the SN phase, i.e. the restframe time since SN peak luminosity:

    .. math::
    
        \mathrm{ph} = \frac{t_{MJD} - t_{max}}{1 + z}

    :math:`M_0`, :math:`M_1` are global surfaces describing the
    spectral evolution of the average SN and its principal variation,
    and :math:`CL(\lambda)` is a global "extinction correction"
    describing the color diversity of the SNIa family.

    :math:`(X_0, X_1, c)` are SN-dependent. :math:`X_0` is the
    amplitude of the "SN-frame B-band lightcurve" as inferred from the
    observer-frame fluxes. :math:`X_1` is the coordinate of the SN
    along :math:`M_1`. In practice, it quantifies the variations of the
    light curve width. :math:`c` is the SN color -- or more exactly,
    the SN color excess with respect to the average SN color.

    Each Light Curve are evaluated interactively to use multi-threading and spectra are evaluated simultaneously.

    Attributes
    ----------
    training_dataset : nacl.dataset.TrainingDataset
        Data set of photometric and spectroscopic observations.
    lc_data : numpy.rec.array
        Photometric data of the training data set.
    phase_range : tuple
        Boundaries for defining the phase interval.
    wl_range : tuple
        Boundaries for defining the wavelength interval.
    n_wl : int
        Number of knot of the wavelength spline basis.
    n_ph : int
        Number of knot of the phase spline basis.
    spectrum_recal_degree : int
        Degree of the spectral recalibration polynomial function.
    init_from_salt2_file : str
        File salt2, created by script 'nacl.simulation.make_salt2_npz.py'.
    delta_phase : float
        Offset in phase wrt SALT2.4.
    dphase_dtmax : numpy.array
        Derivative phase basis wrt time.
    filterpath : str
        Path to the filters files repository in nacl.dev/data/
    timing : list
        Timing of photometric and spectroscopic evaluation.
    basis : nacl.lib.bspline.BSpline2D
        2D spline basis.
    gram : scipy.sparse.csc_matrix
        First order grammian use to evaluate integral product of Wavelength basis of the model and filter basis.
        :math::` \int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda`
    G2 : scipy.sparse.csc_matrix
        Second order grammian use to evaluate integral product of Wavelength basis of the model and filter basis.
        :math::` \int \lambda^2 B_\ell(\lambda) B_q(\lambda) d\lambda`
    L_eff : scipy.sparse.coo_matrix
        Effective wavelength used to calculate the photometric data color law.
    val : numpy.array
        Gather evaluation of each light curve and all spectral data.
    jacobian_val : numpy.array
        Gather non-zero derivatives of each light curve and all spectral data.
    jacobian_i : numpy.array
        Gather data index of non-zero derivatives of each light curve and all spectral data.
    jacobian_j : numpy.array
        Gather parameter index of non-zero derivatives of each light curve and all spectral data.
    filter_basis : nacl.lib.bspline.BSpline
        Filter basis defined common to all the bands.
    color_law : ColorLaw
        Color law class.
    polynome_color_law : array-like
        Photometric color law evaluations.
    jacobian_color_law : scipy.sparse.csr_matrix or None
        Photometric color law matrix of derivatives.
    filter_db : nacl.instruments.FilterDb
        Filter transmission projected on B-splines
    bands : numpy.array
        Bands used for photometric data
    lambda_c : numpy.array
        Mean wavelength of the filter in the SN-restframe.
    lambda_c_red : numpy.array
        Reduced mean wavelength of the filter in the SN-restframe.
    pars : FitParameters
        Model parameters.
    pars0 : numpy.array
        Model initial parameters.
    NORM : float
        Model normalization.
    queue : list
        List of photometric and spectral EvalUnit class.
    disable_cache : object

    """
    
    def __init__(self, training_dataset,
                 phase_range=(-20., 50.), wl_range=(2000., 9000.),  # . 11000.),
                 basis_knot=[127, 20],
                 basis_filter_knots = 900, 
                 spectrum_recal_degree=None,
                 init_from_salt2_file=None,
                 normalization_band_name='SWOPE::B'):
        """Constructor. 
          - instantiate the bases (phase, wavelength, filter)
          - compute the Grams
          - compute the lambda_eff matrix
          - instantiate a color law and evaluate it on the lambda_eff's
          - organizes the data into spectral/photometric computing units         

        Parameters
        ----------
        training_dataset : nacl.dataset.TrainingDataset
            Training dataset.
        phase_range : 2-tuple
            Nominal phase range of the model.
        wl_range : 2-tuple
            Nominal wavelength range of the model.
        basis_knot : 2-list
            Number of wavelength and photometric knot.
        basis_filter_knots : int
            Number of knot for the filter basis
        spectrum_recal_degree : int
            Degree of the spectrum recalibration polynomials
        init_from_salt2_file : str
            If the model is to be initialized with the original
            SALT2-4 surfaces and bases, get them from this file.
        normalization_band_name : str
            Use this band to normalize the model.
        """
        self.training_dataset = training_dataset
        self.lc_data = training_dataset.lc_data
        self.phase_range = phase_range
        self.wl_range = wl_range
        self.n_wl, self.n_ph = basis_knot[0],  basis_knot[1]
        self.spectrum_recal_degree = spectrum_recal_degree
        self.delta_phase = 0.
        self.filterpath = self.training_dataset.filterpath
        
        self.timing = []

        self.basis, self.gram, self.G2, self.L_eff = None, None, None, None
        self.val, self.jacobian_val, self.jacobian_i, self.jacobian_j = None, None, None, None
        self.polynome_color_law, self.jacobian_color_law = None, None

        # Filter database
        # TODO: we should be able to optimize this a little
        # (i.e. adapt the filter bases, to lower the size of the grams)
        self.filter_basis = bspline.BSpline(np.linspace(self.wl_range[0], self.wl_range[1],  # 2000., 9000.,  #
                                                        basis_filter_knots), order=4)  # was 9000
        self.filter_db = FilterDb(self.filter_basis, self.training_dataset,
                                  additional_band=['SWOPE::B'],
                                  filterpath=self.filterpath)
        self.color_law = ColorLaw()
        lc = self.training_dataset.lc_data

        d = dict({(lc[i]['band_id'], lc[i]['Filter']) for i in np.arange(len(lc))})
        self.bands = np.array([d[i] for i in range(len(d))])
        
        # lambda_c needed for the color scatter estimation
        lambda_c = self.training_dataset.lc_data['Wavelength']/(1+self.training_dataset.lc_data['ZHelio'])
        idx_lam_c = np.array([lc.data['i'][0] for lc in self.training_dataset.lcs])
        self.lambda_c = lambda_c[idx_lam_c]
        self.lambda_c_red = self.color_law.reduce(self.lambda_c)
        
        # initialize degree of the spectral recalibration polynomial
        self._init_spectral_polynomial()
        # initialize model basis and model parameter vector 
        if init_from_salt2_file:
            self.init_from_salt2(init_from_salt2_file)
        else:
            # model component: bases, color law, and parameter vector
            self._init_bases()
            # parameter vector 
            self.pars = self.init_pars()
        self.pars0 = self.pars.full.copy()
        # model normalization 
        self.norm = self.normalization(band_name=normalization_band_name)

        # grams
        self.init_grams_and_cc_grid()
        
        # and finally, prepare the computing units
        self.queue = self._init_computing_units()

        # clear the cache
        self.clear_cache()

        # if we need to run queue units separately,
        # could be useful to disable caching of the results
        self.disable_cache = False

    def init_from_salt2(self, salt2_filename):
        """Load the SALT2.4 global surfaces and color law, and adapts them to
        our model implementation. 

        Parameters
        ----------
        salt2_filename : str
            a salt2.npz filename containing the definition of the SALT2.4 bases,
            M0, M1 surfaces and color_law.
        """
        f = np.load(salt2_filename)
        # initialize the original SALT2 basis
        # (just extend it so that it covers [-100,100] in phase)
        #        phase_grid = np.hstack([np.linspace(-100., -21., 5),
        #                                f['phase_grid'],
        #                                np.linspace(55., 100., 5)])
        phase_grid = f['phase_grid']
        wl_grid = f['wl_grid']
        self.basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)
        self.delta_phase = +0.7

        # and init the parameters
        self.pars = self.init_pars()
        self.pars['M0'].full[:] = f['M0'].T.ravel()
        self.pars['M1'].full[:] = f['M1'].T.ravel()
        self.pars['CL'].full[:] = f['CL_pars'][0:4]
        
    def init_grams_and_cc_grid(self):
        r"""Compute the :math:`\Lambda^{\mathrm{eff}}_{\ell q}` matrix (see definition above) on which the color
        law is evaluated.

        .. math::
             \bar{\lambda}_{\ell q} = \frac{\int \lambda^2 B_\ell(\lambda) B_q(\lambda)
             d\lambda}{\int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda}

        Compute de grammian of order one, :math:`G` and two :math:`G2` of the model :

        .. math::
            G = \int \lambda B_\ell(\lambda) B_q(\lambda) d\lambda \\
            G2 = \int \lambda^2 B_\ell(\lambda) B_q(\lambda) d\lambda


        .. note::
             this method will be moved to the main model class (since we may work 
             with one single :math:`\Lambda^{\mathrm{eff}}` matrix in a near future).
        """        
        self.gram = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=1)
        self.G2 = get_gram(0., self.basis.bx, self.filter_basis, lambda_power=2)
            
        gram = self.gram.tocoo()
        gram2 = self.G2.tocoo()
        assert(~np.any(gram.row-gram2.row) and ~np.any(gram.col-gram2.col))
        l_eff = gram2.data / gram.data
        self.L_eff = scipy.sparse.coo_matrix((l_eff, (gram.row, gram.col)), shape=gram.shape)

    def init_pars(self):
        """
        Instantiate a fit parameter vector and return it.

        Returns
        -------
        fp : nacl.lib.fitparameters.FitParameters
            Model parameters.
        """
        nb_sne = self.training_dataset.nb_sne()
        nb_spectra = self.training_dataset.nb_spectra()
        nb_passbands = len(self.bands)
        all_degree = self.recalibration_degree + 1
        nb_lightcurves = self.training_dataset.nb_lcs()
        fp = FitParameters([('X0', nb_sne),                            
                            ('X1', nb_sne),
                            ('c',  nb_sne),
                            ('tmax', nb_sne),
                            ('M0',  self.basis.bx.nj * self.basis.by.nj),
                            ('M1',  self.basis.bx.nj * self.basis.by.nj),
                            ('CL',  4),
                            # TODO: size of recalibration parameters should be abstracted 
                            ('SpectrumRecalibration', all_degree.sum()),
                            ('eta_calib', nb_passbands),
                            ('kappa_color', nb_lightcurves)])

        fp['X0'].full[:] = 1.
        rec_val = np.zeros_like(fp['SpectrumRecalibration'].full[:])
        
        for i in np.arange(nb_spectra):
            offset = all_degree[:i+1].sum()
            rec_val[offset-1] = 1  # -0.2

        fp['SpectrumRecalibration'].full[:] = rec_val  # + 0.2
        fp['kappa_color'].full[:] = 0  # 1e-1
        return fp

    def _init_spectral_polynomial(self):
        """
        Instantiate spectral polynomial for recalibration.
        If : a degree is given, all spectra as the same, which is spectrum_recal_degree
        else : 
        Since the degree of the polynomial must not exceed the number of Light curves :
        this degree is large odd number inferior to the numbers of light curves for this SN.
        """
        nb_spectra = self.training_dataset.nb_spectra()
        self.recalibration_degree = np.ones(nb_spectra).astype(int)
        # print(self.spectrum_recal_degree)
        # print(self.spectrum_recal_degree is None)
        # print(self.spectrum_recal_degree is not None)
        if self.spectrum_recal_degree is None:
            # print(self.spectrum_recal_degree)
            lcs = self.training_dataset.lc_data
            sps = self.training_dataset.spec_data
        
            for isp in np.unique(sps['spec_id']):
                isn = sps['sn_id'][sps['spec_id'] == isp][0]
                n_lcs_sn = len(np.unique(lcs[lcs['sn_id'] == isn]['lc_id']))
                if n_lcs_sn % 2 == 0:
                    self.recalibration_degree[isp] = n_lcs_sn - 1 
                elif n_lcs_sn % 2 == 1:
                    self.recalibration_degree[isp] = n_lcs_sn - 2 
        else:
            self.recalibration_degree *= self.spectrum_recal_degree

    def _init_bases(self):
        """
        Instantiate model bases
        """
        phase_grid = np.hstack([np.linspace(self.phase_range[0], self.phase_range[1], self.n_ph)])
        wl_grid = np.linspace(self.wl_range[0], self.wl_range[1], self.n_wl)
        self.basis = bspline.BSpline2D(wl_grid, phase_grid, x_order=4, y_order=4)

    def normalization(self, band_name='SWOPE::B'):
        r"""Compute the default model normalization.

        The model is normalized so that the broadband flux at peak, of
        a normal :math:`(X1=0, c=0)` SN, in the band specified in
        argument (generally some flavour of the :math:`B`-band) is
        equal to unity.

        .. math::
               \int S(\lambda,0) \frac{\lambda}{hc} B(\lambda) d\lambda = 1

        Returns
        -------
        float
            Overall normalization term of the model.
        """
        tq, filter_basis = self.filter_db[band_name]
        phase_eval = self.basis.by.eval(np.array([0. + self.delta_phase])).tocsr()
        gram = get_gram(z=0., model_basis=self.basis.bx, filter_basis=filter_basis, lambda_power=1)
        surface_0 = self.pars['M0'].full.reshape(len(self.basis.by), -1)
        # evaluate the integral of the model in the specified band
        norm = phase_eval.dot(surface_0.dot(gram.dot(tq)))
        return 1. / norm

    def _init_computing_units(self):
        """Prepare and queue all the computing units

        The model evaluation is deferred to two kinds of so-called
        'eval units'.

            - Photometric eval units, which compute the light curves
              (i.e. the integral of the SALT2 surfaces). In our current
              design, a Photometric eval unit predicts the lightcurve
              for a given SN in a given band.
            - Spectroscopic eval units, which predict SN spectra.
              In our current design, a spectroscopic eval unit
              predicts one spectrum for a given SN.

        All units are grouped in a single execution queue.

        .. note:: we designed things like this because we had some
             hope that we could gain in speed from using explicit
             parallelism (using joblib or some similar framework).  In
             practice, it seems that the overheads from using python
             parallelism at this level are high (because of the GIL)
             and that we gain much more by using numpy/scipy implicit
             multithreading. We stick to this design because it makes
             things clearer in fact.

        """
        queue = []
        for lc in self.training_dataset.lcs:
            queue.append(LightcurveEvalUnitTz(lc, self))
        queue.append(SpectrumEvalUnitFast(self.training_dataset, self))
        return queue

    def update_computing_units(self, ilc=None, spec=False):
        """
        Recreate the model queue as a function of the wanted LCs and spectra.

        Parameters
        -------
        ilc : None or int
            If None all light curve are evaluated, else only teh desired one.
        spec : bool
            Whether Spectra need to be evaluated.

        Returns
        -------
        queue : list
            The new queue
        """
        queue = []
        if ilc is not None:
            for lc in [self.training_dataset.lcs[ilc]]:
                queue.append(LightcurveEvalUnitTz(lc, self))
        if spec:
            queue.append(SpectrumEvalUnitFast(self.training_dataset, self))
        return queue 

    def clear_cache(self):
        """
        Clear model evaluation, derivatives and timing.
        """
        self.val = []
        self.jacobian_i = []
        self.jacobian_j = []
        self.jacobian_val = []
        self.timing = []
        
    def precompute_color_law(self, cl_pars=None, jac=False):
        """
        Precompute the color law for the photometric data.

        Parameters
        -------
        cl_pars : None or numpy.array
            Color law parameters
        jac : bool
            If derivatives are needed.
        """
        if cl_pars is None:
            cl_pars = self.pars['CL'].full
            
        polynome_color_law, jacobian_color_law = self.color_law(self.L_eff.data, cl_pars, jac=jac)  # return_jacobian_as_coo_matrix=False)
        self.polynome_color_law = scipy.sparse.csr_matrix((polynome_color_law,
                                                           (self.L_eff.row, self.L_eff.col)), shape=self.L_eff.shape)
        if jac:
            self.jacobian_color_law = []
            _, n = jacobian_color_law.shape
            for i in range(n):
                jacobian_cl = scipy.sparse.csr_matrix((jacobian_color_law[:, i],
                                                       (self.L_eff.row, self.L_eff.col)), shape=self.L_eff.shape)
                self.jacobian_color_law.append(jacobian_cl)
    
    def get_restframe_phases(self, data):
        """
        Given a chunk of data, return the restframe phases

        Returns
        -------
        numpy.array
            data SN-restframe phase
        """
        sn_id = data['sn_id']
        tmax = self.pars['tmax'].full[sn_id]
        z = self.training_dataset.sne['z'][sn_id]
        return (data['Date'] - tmax) / (1.+z)
                
    def __call__(self, p, jac=False, plotting=False, ilc_plot=None, spec_plot=False):
        """Evaluate the model for the parameter set p 

        In practice, loop over the queue and run each eval unit.
        Assemble the results into one single vector and (optionally)
        one single jacobian matrix.

        Parameters
        -------
        p : numpy.array
            Vector containing the free parameters only.
        jac : bool
            Whether it return the jacobian matrix.

        plotting : bool
            Plotting need to evaluated only some data [plotting needs]
        ilc_plot : list
            Light curves to be evaluated [plotting needs]
        spec_plot : bool
            Spectra to be evaluated [plotting needs]


        Returns
        -------
        val : numpy.array
            model results
        jacobian : scipy.sparse.csr_matrix
            jacobian matrix (if jac is true)
        """
        # update pars
        self.pars.free = p

        # pre-evaluate the color law (shared between all light curves)
        self.precompute_color_law(jac=jac)

        # precompute the values of the phase basis on the full dataset 
        # significantly faster that way
        restframe_phases = self.get_restframe_phases(self.lc_data)
        self.phase_eval = self.basis.by.eval(restframe_phases + self.delta_phase).tocsr()
        if jac:
            self.dphase_dtmax = self.basis.by.deriv(restframe_phases + self.delta_phase).tocsr()
        else:
            self.dphase_dtmax = None
            
        self.clear_cache()

        # queue all Eval Unit for minimization or for plotting
        if plotting:
            queue = self.update_computing_units(ilc=ilc_plot, spec=spec_plot)
        else:
            queue = self.queue
            
        # now, we are ready to loop over the eval units 
        for q in queue:
            q(jac)
        val = np.hstack(self.val)
        
        if not jac:
            return val

        # collect and assemble the results 
        n_data = len(self.training_dataset.lc_data)  # + len(self.training_dataset.spec_data)
        fit_spec = False
        for ique in self.queue:
            if ique.data['spec_id'].mean() != -1:
                fit_spec = True
                
        if fit_spec:
            n_data += len(self.training_dataset.spec_data)
            
        n = len(self.pars.free)
        
        i = np.hstack(self.jacobian_i)
        j = np.hstack(self.jacobian_j)
        v = np.hstack(self.jacobian_val)
        idx = j >= 0  # self.pars.indexof(j) >= 0
        jacobian = scipy.sparse.coo_matrix((v[idx], (i[idx], j[idx])), shape=(n_data, n))
        return val, jacobian

#       else:
#            logging.info('SALT2Like.__call__: joblib.Parallel')
#            def f(x):
#                x.__call__(jac=jac)
#            Parallel(n_jobs=n_jobs)(delayed(f)(x) for x in self.queue)
#        self.assemble(res)

# move this into model (cache needs to be associated with the bases)


_G, _G2 = {}, {}


def get_gram(z, model_basis, filter_basis, lambda_power=1):
    """
    Calculate the grammian of to spline basis.

    The grammian :math:`G` of order :math:`N`, to basis :math:`B_0` (define on wavelength,
    in our case surfaces wavelength basis) and
    :math:`B_1` (define on SN-restframe wavelength, filter basis) is defined as :

    .. math::
        G = \int \lambda^N B_0(\lambda) B_q(\lambda (1+z)) d\lambda

    Parameters
    -------
    z : numpy.array
        Vector containing the data redshift.
    model_basis : nacl.lib.bspline.BSpline
        Wavelength basis.
    filter_basis : nacl.lib.bspline.BSpline
        Filter basis defined common to all the bands.
    lambda_power : int
        Grammian order.

    Returns
    -------
    gram : scipy.sparse.csc_matrix
        Grammian
    """
    global _G
    global _G2
    key = (z, model_basis, filter_basis, lambda_power)
    
    gram = _G.get(key, None)
    if gram is None:
        gram = bspline.lgram(model_basis, filter_basis, z=z, lambda_power=lambda_power).tocsc()
        _G[key] = gram
        
    return gram


class LightcurveEvalUnitTz(object):
    r"""
    Compute an SN light curve of one supernova in one single band.

    This class is one of the two types of "computing unit". Given a
    set of dates, and the :math:`(X_0, X_1, c, t_{max})` parameters of
    the SN, compute the quantities:

    .. math::

         X_0 \times (1+z) \times \left[I_0 + X_1 I_1\right]

    with

    .. math::
         I_{0|1} = \int M_{0|1}(\lambda, \mathrm{p}) \frac{\lambda}{hc} T((1+z)\lambda) 10^{0.4\ c\ P(\lambda)} d\lambda

    In practice, we can speed up the computations by decomposing the
    filter on a spline basis: :math:`T(\lambda (1+z)) = \sum_q t_q(z)
    B_q(\lambda)= \sum_q (1+z) t_q B_q(\lambda)`, which allows to decompose the integral above as:

    .. math::
        I_{0|1} = \sum_{k\ell q} \theta_{0|1, k\ell} t_q(z) B_k(\mathrm{p})\ \int B_\ell(\lambda)
        \frac{\lambda}{hc} B_q(\lambda)\ 10^{0.4\ c\ P(\lambda)} d\lambda

    We use the fact that the color law is a slowly variable function
    of :math:`\lambda` (compared to the individual splines) to expel the
    color law from the integral:

    .. math::
       \int B_\ell(\lambda) \frac{\lambda}{hc} B_q(\lambda) CL(\lambda) d\lambda \approx
       CL(\lambda^{\mathrm{eff}}_{\ell q})\ \times \int B_\ell(\lambda) \frac{\lambda}{hc} B_q(\lambda) d\lambda

    up to second order, if we define:

    .. math::
       \lambda^{\mathrm{eff}}_{\ell q} = \frac{\int B_\ell(\lambda)
       \lambda^2 B_q(\lambda) d\lambda}{\int B_\ell(\lambda) \lambda B_q(\lambda) d\lambda}


    and so, defining: :math:`\Lambda^{\mathrm{eff}} = (\lambda^{\mathrm{eff}}_{\ell q})` and
    ::math:`G = (G_{\ell q}) = (\int B_\ell(\lambda)\frac{\lambda}{hc} B_q(\lambda)) d\lambda`,
    the integral evaluation reduces to:

    .. math::
        I_{0|1} = J\ \cdot \Theta\ \cdot \left[G \Lambda^{\mathrm{eff}}\right] \cdot t_z

    where :math:`J`, :math:`G`,
    :math:`\Lambda^{\mathrm{eff}}` and :math:`t_z` are sparse
    and where :math:`G, \Lambda^{\mathrm{eff}}`
    can be precomputed once for all.

    Matrix :math:`G` is called the Gramian of the model and
    filter spline bases. It is precomputed exactly, using Gaussian
    quadrature when the model is instantiated.

    This strategy decompose :math:`T((1+z)\lambda)` and work with a single Gramian and
    an SN-dependant :math:`t_z` vector.
    This option is probably computationally less intensive, as we can
    factorize the evaluation of the color law.

    Attributes
    ----------
    lcdata : nacl.dataset.LcData
        Photometric of one Lc of one SN nacl data class.
    data : numpy.rec.array
        Photometric data in nacl data file type.
    model : nacl.models.salt.SALT2Like
        Model.
    z : numpy.array
        SN redshift of Lc observation.
    band : numpy.array
         SN band of Lc observation.
    tqz : numpy.array
         :math:`t_z`.
    sn_id : numpy.array
        SNe index of all data.
    ph_basis_size ; int
        Number of phase basis knots.
    M0 : numpy.array
        M_0 parameters reshape as 2D array, (wavelength, phase) matrix.
    M1 : numpy.array
        M_1 parameters reshape as 2D array, (wavelength, phase) matrix.
    cl_pars : numpy.array
        Color law parameters.
    """

    def __init__(self, lcdata, model):
        r"""Constructor
        Retrieve the band shape :math:`T(\lambda)` and project :math:`T(\lambda(1+z))` on the basis

        Parameters
        ----------
        lcdata : nacl.dataset.LcData
            Photometric of one Lc of one SN nacl data class.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.lcdata = lcdata
        self.data = lcdata.data
        self.model = model
        self.z = lcdata.z        
        self.band = lcdata.band
        tr = model.filter_db.transmission_db[self.band]
        self.tqz, _ = model.filter_db.insert(tr, z=self.z)
        # set to zero very small values
        self.tqz[(np.abs(self.tqz)/self.tqz.max() < 1e-10)] = 0.
        
        self.sn_id = lcdata.sn_id
        
        #        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)
        
        self.M0 = model.pars['M0'].full.reshape(self.ph_basis_size, -1)
        self.M1 = model.pars['M1'].full.reshape(self.ph_basis_size, -1)
        self.cl_pars = model.pars['CL'].full

    def __call__(self, jac=False, debug_mode=False):
        """
        Evaluate the SN light curve.

        Parameters
        ----------
        jac : bool
            Whether it return the jacobian matrix
        debug_mode : bool
            if True, return the model components instead of the
            model results.

        Returns
        ----------
        val : numpy.array
            Model evaluation
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (optional)

        """
        t0 = time.time()
        sn_id = self.sn_id
        zz = 1. + self.z
        pars = self.model.pars
        data = self.lcdata.data
        band_id = self.lcdata.data['band_id'][0]
        # print(self.model.bands[band_id], self.lcdata.data['Filter'][0])
        lc_id = self.lcdata.data['lc_id'][0]

        # sn-related parameters
        x0, x1 = pars['X0'].full[sn_id], pars['X1'].full[sn_id]
        c, tmax = pars['c'].full[sn_id],  pars['tmax'].full[sn_id]
        eta = pars['eta_calib'].full[band_id]

        kappa = pars['kappa_color'].full[lc_id]
        
        phase_eval = self.model.phase_eval[self.lcdata.slc]
        # color law evaluation -- on the Gram \lambda_eff's
        cl = np.power(10., 0.4 * c * self.model.polynome_color_law.data)
        csr_color_law = scipy.sparse.csr_matrix((cl, (self.model.L_eff.row, self.model.L_eff.col)),
                                                shape=self.model.L_eff.shape)
        gram = self.model.gram.multiply(csr_color_law)
        tqz = self.tqz

        # much faster with the .dot's (instead of @)
        integral_surface_0 = phase_eval.dot(self.M0.dot(gram.dot(tqz)))
        integral_surface_1 = phase_eval.dot(self.M1.dot(gram.dot(tqz)))
        flux = self.model.norm * zz * x0 * (integral_surface_0+x1*integral_surface_1)*(1+eta)*(1+kappa)

        if debug_mode:
            return self.model.norm * integral_surface_0, self.model.norm * integral_surface_1, \
                   cl, flux, x0, x1, self.model.norm

        self.model.val.append(flux)
        if not jac:
            self.model.timing.append(time.time()-t0)
            return flux

        n_data = len(self.data)
        sn = np.full(n_data, self.sn_id)
        norm = self.model.norm
        
        # shortcut names to the internal cache holding the
        # jacobian matrix definition 
        jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        
        # dMdX0
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['X0'].indexof(sn))
        jacobian_val.append(norm * zz * (integral_surface_0 + x1*integral_surface_1) * (1+eta) * (1+kappa))

        # dMdX1
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['X1'].indexof(sn))
        jacobian_val.append(norm * zz * x0 * integral_surface_1 * (1+eta) * (1+kappa))

        # dMdc
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['c'].indexof(sn))
        gram_color_law = gram.multiply(self.model.polynome_color_law)
        dintegral0_dc = 0.4 * np.log(10.) * phase_eval.dot(self.M0.dot(gram_color_law.dot(tqz)))
        dintegral1_dc = 0.4 * np.log(10.) * phase_eval.dot(self.M1.dot(gram_color_law.dot(tqz)))
        jacobian_val.append(norm * zz * x0 * (dintegral0_dc + x1*dintegral1_dc) * (1+eta) * (1+kappa))

        # dMdtmax
        dphase_dtmax = self.model.dphase_dtmax[self.lcdata.slc]
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['tmax'].indexof(sn))
        dintegral0_dtmax = -dphase_dtmax.dot(self.M0.dot(gram.dot(tqz))) / zz
        dintegral1_dtmax = -dphase_dtmax.dot(self.M1.dot(gram.dot(tqz))) / zz
        jacobian_val.append(norm * zz * x0 * (dintegral0_dtmax + x1*dintegral1_dtmax)*(1+eta)*(1+kappa))

        # dMdtheta_0
        dbase_dtheta = scipy.sparse.kron(phase_eval, gram.dot(tqz)).tocoo()
        jacobian_i.append(self.data.i[dbase_dtheta.row])
        jacobian_j.append(self.model.pars['M0'].indexof(dbase_dtheta.col))
        jacobian_val.append(norm * zz * x0 * dbase_dtheta.data * (1+eta) * (1+kappa))
        
        # dMdtheta_1
        jacobian_i.append(self.data.i[dbase_dtheta.row])
        jacobian_j.append(self.model.pars['M1'].indexof(dbase_dtheta.col))
        jacobian_val.append(norm * zz * x0 * x1 * dbase_dtheta.data * (1+eta) * (1+kappa))
        
        # dMdcl
        # I really don't know how to vectorize the computation
        # of these derivatives. If somebody has an idea let me know. 
        buff = np.zeros(n_data).astype(int)
        for i, jacobian_color_law in enumerate(self.model.jacobian_color_law):
            buff[:] = i
            jacobian_i.append(self.data.i)
            jacobian_j.append(self.model.pars['CL'].indexof(buff))
            gram_jacobian_color_law = gram.multiply(jacobian_color_law)
            jacobian_surface0_color_law = 0.4 * np.log(10.) * c * phase_eval.dot(self.M0.dot(
                gram_jacobian_color_law.dot(tqz)))
            jacobian_surface1_color_law = 0.4 * np.log(10.) * c * phase_eval.dot(self.M1.dot(
                gram_jacobian_color_law.dot(tqz)))
            jacobian_val.append(norm * zz * x0 * (jacobian_surface0_color_law + x1*jacobian_surface1_color_law)
                                * (1+eta) * (1+kappa))

        # dMdeta
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['eta_calib'].indexof(self.lcdata.data['band_id']))
        jacobian_val.append(norm * x0 * zz * (integral_surface_0 + x1*integral_surface_1)*(1+kappa))

        # dKappa
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['kappa_color'].indexof()[data['lc_id']])
        jacobian_val.append(norm * x0 * zz * (integral_surface_0 + x1*integral_surface_1) * (1+eta))
    
        if self.model.disable_cache:
            self.model.clear_cache()
        self.model.timing.append(time.time()-t0)
        return flux


class SpectrumEvalUnitFast(object):
    r"""Evaluate the model for all SN spectra in the training dataset
    
    This class is one of the two type of "eval units". Given a chunk
    of the training dataset which corresponds the spectral observations,
    and given the :math:`(X_0, X_1, c, t_{max})` parameters 
    of the SN, compute the quantity:
    
    .. math::
    
         \frac{1}{1+z} \left[M_0\left(\frac{\lambda}{1+z}, \mathrm{p}\right) + X_1\ M_1\left(\frac{\lambda}{1+z},
         \mathrm{p}\right) \right]\ 10^{0.4\ c\ P(\frac{\lambda}{1+z})}\ s(\lambda_{rec})

    where

    .. math::
        M_{0|1}(\lambda, \mathrm{p}) = \sum_{k\ell} \theta_{k\ell} B_k(\mathrm{p}) B_l(\mathrm{\lambda}) 

    and 

    .. math::
         \mathrm{p} = \frac{t - t_{max}}{1+z}

    and where :math:`R_s(\lambda)` is a polynomial correction which
    absorbs the wavelength-dependent large scale calibration errors
    affecting the spectrum.
    
    Again, the evaluation reduces to a sparse matrix multiplication.

    Attributes
    ----------
    training_dataset : nacl.dataset.TrainingDataset
        Data set of photometric and spectroscopic observations.
    data : nacl.dataset.SpectrumData
        Spectra of all SNe in the training dataset.
    model : nacl.models.salt.SALT2Like
        Model.
    z : numpy.array
        SN redshift of Lc observation.
    spec_id : numpy.array
        spectrum index of all data.
    sn_id : numpy.array
        SNe index of all data.
    ph_basis_size : int
        Number of phase basis knots.
    wl_basis_size : int
        Number of wavelength basis knots.
    restframe_wl : numpy.array
        SNe-restframe wavelength for all spectra.
    recal_func : SpectrumRecalibrationPolynomials
        Recalibration polynomial.
    M0 : numpy.array
        M_0 parameters reshape as 2D array, (wavelength, phase) matrix.
    M1 : numpy.array
        M_1 parameters reshape as 2D array, (wavelength, phase) matrix.
    cl_pars : numpy.array
        Color law parameters.

    """
    def __init__(self, tds, model):
        """Constructor.

        Parameters
        ----------
        tds : nacl.dataset.TrainingDataset
            Data set of photometric and spectroscopic observations.
        model : nacl.models.salt.SALT2Like
            Model.
        """
        self.training_dataset = tds
        self.data = tds.sp_data
        self.model = model

        self.spec_id = self.data['spec_id']
        self.sn_id = self.data['sn_id']
        self.z = self.training_dataset.sne['z'][self.sn_id]
        
        # Look at that later 
        self.color_law = model.color_law
        self.pars = model.pars
        self.basis = model.basis
        
        self.wl_basis_size = len(model.basis.bx)
        self.ph_basis_size = len(model.basis.by)
        
        # restframe wavelengths
        self.restframe_wl = self.data['Wavelength']/(1.+self.z)
        #        self.Jl = model.basis.bx.eval(self.restframe_wl).tocsr()

        # and we can connect directly to the global parameters 
        self.M0 = model.pars['M0'].full
        self.M1 = model.pars['M1'].full
        self.cl_pars = model.pars['CL'].full
        
        # recalibration polynomial (one per spectrum, because we adapt to the spectrum wavelength range)
        self.recal_func = SpectrumRecalibrationPolynomials(self.training_dataset, self.model,
                                                           self.model.recalibration_degree)

    def __call__(self, jac=False, debug_mode=False):
        r"""
        Evaluate the model for a all spectra

        Parameters
        ----------
        jac : bool
            if True compute and return the jacobian matrix
        debug_mode : bool
            if true, just return the model components.

        Returns
        -------
        val : numpy.array
            Model evaluations.
        jacobian : scipy.sparse.csr_matrix
            Jacobian matrix (if jac is true).
        """
        t0 = time.time()
        pars = self.pars

        # sn-related parameters
        x0, x1 = pars['X0'].full[self.sn_id], pars['X1'].full[self.sn_id]
        c, tmax = pars['c'].full[self.sn_id],  pars['tmax'].full[self.sn_id]

        # we need to re-evaluate the basis on the phases, since tmax changes
        restframe_phases = (self.data['Date']-tmax)/(1.+self.z)
        jacobian = self.basis.eval(self.restframe_wl, restframe_phases + self.model.delta_phase).tocsr()
        
        # model components
        component_0 = jacobian.dot(self.M0)
        component_1 = jacobian.dot(self.M1)
        polynome_color_law, jacobian_color_law = self.color_law(self.restframe_wl, self.cl_pars, jac=jac)
        color_law = np.power(10., 0.4*c*polynome_color_law)
        zz = 1. + self.z
        
        # recalibration polynomial
        recal, jacobian_spec_rec = self.recal_func(jac=jac)
        
        # recal = np.exp(recal)
        pca = (component_0 + x1 * component_1)
        model = pca * color_law * recal / zz
        
        if debug_mode:
            return component_0, component_1, color_law, recal, model
        
        self.model.val.append(model)
        if not jac:
            self.model.timing.append(time.time()-t0)
            return model

        jacobian = jacobian.tocoo()
        jacobian_spec_rec = jacobian_spec_rec.tocoo()
        # jacobian_spec_rec.data *= recal[jacobian_spec_rec.row]
        # X0 does not appear in the spectral part of the model 
        # hence, dmdX0 = 0

        # dMdX1
        jacobian_i, jacobian_j, jacobian_val = self.model.jacobian_i, self.model.jacobian_j, self.model.jacobian_val
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['X1'].indexof(self.sn_id))
        jacobian_val.append(component_1 * color_law * recal / zz)

        # dMdc
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['c'].indexof(self.sn_id))
        jacobian_val.append(model * 0.4 * np.log(10.) * polynome_color_law)

        # dMdtmax
        # we can gain a little here, by not evaluating the gradient along the wavelength (ddlambda)
        _, deval_phase = self.model.basis.gradient(self.restframe_wl, restframe_phases + self.model.delta_phase)
        deval_phase = deval_phase.tocsr()
        jacobian_i.append(self.data.i)
        jacobian_j.append(self.model.pars['tmax'].indexof(self.sn_id))
        jacobian_val.append(-1. * (deval_phase.dot(self.M0) + x1*deval_phase.dot(self.M1)) * color_law * recal / zz**2)

        # dmdtheta_0
        jacobian_i.append(self.data.i[jacobian.row])
        jacobian_j.append(self.model.pars['M0'].indexof(jacobian.col))
        jacobian_val.append(jacobian.data * color_law[jacobian.row] * recal[jacobian.row] / zz[jacobian.row])

        # dmdtheta_1
        jacobian_i.append(self.data.i[jacobian.row])
        jacobian_j.append(self.model.pars['M1'].indexof(jacobian.col))
        jacobian_val.append(x1[jacobian.row] * jacobian.data * color_law[jacobian.row] *
                            recal[jacobian.row] / zz[jacobian.row])
                
        # dMdcl (color law)
        jacobian_color_law = scipy.sparse.coo_matrix(jacobian_color_law)
        jacobian_i.append(self.data.i[jacobian_color_law.row])
        jacobian_j.append(self.model.pars['CL'].indexof(jacobian_color_law.col))
        jacobian_val.append(c[jacobian_color_law.row] * 0.4 * np.log(10.) * jacobian_color_law.data *
                            model[jacobian_color_law.row])
        
        # dMdr (recalibration)
        jacobian_i.append(self.data.i[jacobian_spec_rec.row])
        jacobian_j.append(self.model.pars['SpectrumRecalibration'].indexof(jacobian_spec_rec.col))
        jacobian_val.append(jacobian_spec_rec.data * (pca*color_law)[jacobian_spec_rec.row]/zz[jacobian_spec_rec.row])
        
        if self.model.disable_cache:
            self.model.clear_cache()
        self.model.timing.append(time.time()-t0)
        
        return model



"""
"""
import numpy as np
import logging
import pylab as pl
from . import instruments
logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',
                    level=logging.INFO)


def stringify(nt):
    """
    Cast a record array so that byte fields become utf8 fields
    """
    dt = [(f, t.replace('S', 'U')) for f, t in nt.dtype.descr]
    return nt.astype(dt)


def load(filename):
    """
    default loader: load the simulated data 
    in the format G.A. has defined for his work.

    Parameters
    -------
    filename : str
        name of the simulation file

    Returns
    -------
    TrainingDataset
       A training dataset, with (1) LC and spectra, (2) an index of each spectrum and each light curve
    """
    logging.info('loading data from: {}'.format(filename))
    f = np.load(filename, allow_pickle=True)
    data = f['data']
    sne = f['snInfo']
    
    idx_spectral_data = data['Filter'] == b''
    lc_data = data[~idx_spectral_data]
    spectral_data = data[idx_spectral_data]

    lc_data = np.lib.recfunctions.rename_fields(lc_data, {'id': 'sn_id', 'obs_id': 'spec_id'})
    spectral_data = np.lib.recfunctions.rename_fields(spectral_data, {'id': 'sn_id', 'obs_id': 'spec_id'})
    lc_data['spec_id'] = -1
    lc_data, spectral_data = indexing(lc_data, spectral_data)
    return TrainingDataset(lc_data, spectral_data, sne)


def indexing(lc_data, spectral_data):
    """
    For photometric data create a band index, a light curve index and a data point index.
    For spectroscopic data create a band index (equal to -1), a light curve index (equal to -1)
    and a data point index (starting after last photometric data point index).

    Parameters
    -------
    lc_data : nacl.dataset.LcData
        Photometric data
    spectral_data : nacl.dataset.SpectrumData
        Spectral data

    Returns
    -------
    lc_data : numpy.rec.array (=nacl.dataset.LcData.data)
        Photometric data array.
    spectral_data : numpy.rec.array (=nacl.dataset.SpectrumData.data)
        Spectral data array.
    """
    n_lc, n_sp = len(lc_data), len(spectral_data)

    # band indexation                                                                                
    dict_bd = {}
    _, idx_bd = np.unique(lc_data['Filter'], return_index=True)
    for i_bd, bd in enumerate(lc_data['Filter'][np.sort(idx_bd)]):
        dict_bd[bd] = i_bd
    id_bd = np.array([dict_bd[bd] for bd in lc_data['Filter']])

    # light curve indexation                                                                         
    c = 0
    id_lc = np.ones(len(lc_data['sn_id'])).astype('<i8')
    
    for i in range(lc_data['sn_id'][-1]+1):
        idx_sn = lc_data['sn_id'] == i
        lcs = lc_data[idx_sn]
        _, idx = np.unique(lcs["Filter"], return_index=True)
        for bd_sn in lcs['Filter'][np.sort(idx)]:
            id_lc[(lc_data['sn_id'] == i) & (lc_data['Filter'] == bd_sn)] = c 
            c += 1
            
    id_lc = np.hstack(np.array(id_lc))

    lc_data = np.lib.recfunctions.rec_append_fields(lc_data,
                                                    names=['lc_id', 'band_id', 'i'],
                                                    data=[id_lc, id_bd, np.arange(n_lc)])
    i_sp = n_lc + np.arange(n_sp)
    sp_ones = np.ones_like(i_sp)
    spectral_data = np.lib.recfunctions.rec_append_fields(spectral_data,
                                                          names=['lc_id', 'band_id', 'i'],
                                                          data=[-1*sp_ones, -1*sp_ones,
                                                                n_lc + np.arange(n_sp)])
    return lc_data, spectral_data


class LcData:
    r"""
    Light curve data class.
    We define one per Lc.

    Attributes
    ----------
    sn : int
        Index of sn.
    band : str
        Name of the Filter.
    slc : slc
        Index of the Light Curve data in the full photometric data.
    lc_data : numpy.rec.array
        Full photometric data.
    sne : numpy.rec.array
            Information of all SNe (:math:`(z, X_0, X_1, c, t_{max})`)
    sn_info : numpy.rec.array
        Information of the SN.

    """
    def __init__(self, sn, band, slc, lc_data, sne):
        """
        Constructor

        Parameters
        ----------
        sn : int
            Index of sn.
        band : str
            Name of the Filter.
        slc : slc
            Index of the Light Curve data in the full photometric data.
        lc_data : numpy.rec.array
            Photometric data.
        sne : numpy.rec.array
            Information of all SNe (:math:`(z, X_0, X_1, c, t_{max})`)
        """
        self.sn = sn
        self.band = band.astype(str)
        self.slc = slc
        self.lc_data = lc_data
        self.sne = sne
        self.sn_info = self.sne[self.sn]

    def __len__(self):
        """
        Return number of light curve data point.
        """
        return len(self.lc_data[self.slc])

    @property
    def data(self):
        """
        Return numpy.rec.array data file.
        """
        return self.lc_data[self.slc]

    @property
    def z(self):
        """
        return SNe redshift.
        """
        return self.sn_info['z']

    @property
    def sn_id(self):
        """
        Return SN index.
        """
        return self.sn

    def plot(self):
        """
        control plot
        """
        pl.figure()
        x, y, ey = self.data['Date'], self.data['Flux'], self.data['FluxErr']
        pl.errorbar(x, y, yerr=ey, ls='', color='k', marker='o')
        pl.xlabel('phase [days]')
        pl.ylabel('Flux')
        pl.title('SN#{} {} [$z={:5.3}]$'.format(self.sn_id, self.band, self.z))

        
class SpectrumData:
    r"""
    Spectrum data class.

    Attributes
    ----------
    sn : int
        Index of sn.
    spectrum : str
        Spectra index .
    slc : slc
        Index of the Light Curve data in the full photometric data.
    spec_data : numpy.rec.array
        Full spectral data.
    sne : numpy.rec.array
            Information of all SNe (:math:`(z, X_0, X_1, c, t_{max})`)
    sn_info : numpy.rec.array
        Information of the SN.
    """
    def __init__(self, sn, spectrum, slc, spec_data, sne):
        """
        Constructor

        Parameters
        ----------
        sn : int
            Index of sn.
        spectrum : str
            Spectra index .
        slc : slc
            Index of the Light Curve data in the full photometric data.
        spec_data : numpy.rec.array
            Full spectral data.
        sne : numpy.rec.array
            Information of all SNe (:math:`(z, X_0, X_1, c, t_{max})`)
        """
        self.sn = sn
        self.spectrum = spectrum
        self.slc = slc
        self.spec_data = spec_data
        self.sne = sne
        self.sn_info = self.sne[self.sn]
        
    def __len__(self):
        """
        Return number of spectral point.
        """
        return len(self.spec_data[self.slc])

    @property
    def data(self):
        """
        Return data numpy.rec.array
        """
        return self.spec_data[self.slc]

    @property
    def z(self):
        """
        Return redshift.
        """
        return self.sn_info['z']

    @property
    def sn_id(self):
        """
        Return SN index
        """
        return self.sn
                 
    def plot(self):
        """
        control plot
        """
        pl.figure()
        x, y, ey = self.data['Wavelength'], self.data['Flux'], self.data['FluxErr']
        pl.errorbar(x, y, yerr=ey, ls='', color='k')
        pl.xlabel(r'$\lambda [\AA]$')
        pl.ylabel('Flux')
        pl.title('SN#{} [$z={:5.3}$]'.format(self.sn_id, self.z))
        

class TrainingDataset:
    """
    A class to make sure that the format of the training dataset 
    is hidden from the model. We don't want to change the model 
    each time we change the format.

    Attributes
    ----------
    lc_data : numpy.rec.array
        Light curve data.
    spec_data : numpy.rec.array
        Spectral data.
    sne : numpy.rec.array
        Information of all SNe (:math:`(z, X_0, X_1, c, t_{max})`)
    filter_names : list
        Filters names used in the training dataset.
    transmissions : list
        Filters transmission.
    lcs : list[LcData]
        List of each LcData.
    spectra : list[SpectrumData]
        List of the SpectrumData.
    filterpath : str
        If filterpath should be explicit.
    """
    def __init__(self, lc_data, spec_data, sne, filterpath=None, spectra=None, lcs=None):
        """
        load the file, sort by data type (lc/spec) and SN

        Parameters
        ----------
        lc_data : numpy.rec.array
            Light curve data.
        spec_data : numpy.rec.array
            Spectral data.
        sne : numpy.rec.array
            Information of all SNe (:math:`(z, X_0, X_1, c, t_{max})`)
        lcs : list
            List of each LcData.
        spectra : list
            List of the spectral data.
        filterpath :str
            If filterpath should be explicit.
        """
        self.lc_data = lc_data
        self.spec_data = spec_data
        self.sne = sne

        if lcs is None:
            self.lcs = []
            self._index_lcs()
        else:
            self.lcs = lcs
        
        if spectra is None: 
            self.spectra = []
            self._index_spectra()            
        else:
            self.spectra = spectra
        self.filterpath = filterpath

    def _index_lcs(self):
        """
        Sort the light curve data, so that the light curve measurements
        are stored in continuous chunks and can be indexed easily.
        
        Build and index for each light curve (i.e. set of photometric
        points, for one sn, measured in one band). An index entry
        contains a slice object and links to the original data.
        """
        logging.info('indexing the light curves')

        self.lcs = []
        
        self.lc_data.sort(order=['sn_id', 'lc_id', 'Date'])
        self.lc_data['i'] = np.arange(len(self.lc_data)).astype(int)
        for i_lc in np.arange(self.lc_data['lc_id'][-1]+1):
            idx = (self.lc_data['lc_id'] == i_lc)
            band = self.lc_data['Filter'][idx][0]
            sn = self.lc_data['sn_id'][idx][0]
            i = np.where(idx)[0]
            slc = slice(i.min(), i.max()+1)
            self.lcs.append(LcData(sn, band, slc, self.lc_data, self.sne))

    def _index_spectra(self):
        """
        Sort the spectral data, so that the spectrum measurements are
        stored in contiguous chunks and can be indexed easily.

        Build and index for each spectrum. An index entry contains a
        slice object and links to the original data.
        """
        logging.info('indexing the spectra')
        self.spec_data.sort(order=['sn_id', 'spec_id', 'Wavelength'])
        
        nspin = self.spec_data['spec_id']-self.spec_data['spec_id'][0]
        nspin[np.where(nspin < 0)] += len(np.unique(self.spec_data['spec_id']))
        self.spec_data['spec_id'] = nspin
        
        self.spec_data['i'] = len(self.lc_data) + np.arange(len(self.spec_data)).astype(int)        
                    
    def nb_sne(self):
        """
        Return number of SNe.
        """
        return len(self.sne)

    def nb_lcs(self):
        """
        Return number of light curves.
        """
        return len(self.lcs)

    def nb_spectra(self):
        """
        Return number of spectra.
        """
        n = len(np.unique(self.spec_data['spec_id']))
        return n  # len(self.spectra)

    @property
    def sp_data(self):
        """
        Return numpy.rec.array spectra data.
        """
        return self.spec_data

    @property
    def lcs_data(self):
        """
        Return numpy.rec.array photometric data.
        """
        return self.lc_data

    def get_all_filter_names(self):
        """
        scan the dataset and return the list of all passbands used to take
        the data.

        Returns
        ----------
        list
            List of filter name.
        """
        if not hasattr(self, 'filter_names'):
            t = np.unique(self.lc_data['Filter'])
            self.filter_names = t.astype(t.dtype.str.replace('S', 'U'))
        return self.filter_names

    def get_all_transmissions(self):
        """
        Return transmission of all filters

        Returns
        ----------
        list
            List of filter transmission.

        """
        if hasattr(self, 'transmissions'):
            return self.transmissions
        self.transmissions = [instruments.load(fn,path=self.filterpath) for fn in self.get_all_filter_names()]
        return self.transmissions
    
    def __iter__(self):
        """
        """
        pass



    



import numpy as np
import pylab as pl
import scipy 
import os

ALPHA = 0.13
BETA = 3.0
M_B = -19.0906


class DistancesModulus(object):
    r"""
    Calculate distance modulus from parameter of a nacl.model.salt model.
    Given the parameters of the standardization laws (:math:`\alpha`, :math:`\beta` and :math:`M_B`)
    and the SN parameters from the models (:math:`X_0`, :math:`X_1` and :math:`c`), calculated
    the distance modulus :math:`\mu` :

    .. math::
        \mu = -2.5 \log_{10}(X_{0}) + \alpha X_{1} - \beta c - M_B

    Attributes
    ----------
    model : nacl.models.salt
        Model.
    beta0 : None or numpy.array
        Model initial parameters.
    nsn : int
        Number of SNe in data set.
    """
    def __init__(self, model, beta0=None):
        """
        Constructor.

        Parameters
        ----------
        model : nacl.models.salt
            Model.
        beta0 : numpy.array
            Model initial parameters [optional]
        """
        self.model = model
        self.beta0 = beta0
        self.nsn = model.training_dataset.nb_sne()

    def __call__(self, beta, jac=False):
        """
        Evaluate the distance with parameter :math:`\beta`.

        Parameters
        -------
        beta : numpy.array
            Vector containing the free parameters only.
        jac : bool
            Whether it return the jacobian matrix.

        Returns
        -------
        mu : numpy.array
            distance modulus
        """
        idxx0 = self.model.pars['X0'].indexof()
        idxx1 = self.model.pars['X1'].indexof()
        idxc = self.model.pars['c'].indexof()         
        x0, x1, c = beta[idxx0], beta[idxx1], beta[idxc]
        
        mu = - 2.5 * np.log10(x0) + ALPHA * x1 - BETA * c - M_B
        
        if jac:
            dmu_dx0 = -2.5 * 1/(x0 * np.log(10))
            dmu_dx1 = +ALPHA * np.ones_like(idxx1)
            dmu_dc = - BETA * np.ones_like(idxc)
            i, j, v = [], [], []
            # dx0
            i.append(np.arange(self.nsn))
            j.append(idxx0)
            v.append(dmu_dx0)
            # dx1
            i.append(np.arange(self.nsn))
            j.append(idxx1)
            v.append(dmu_dx1)
            # dc
            i.append(np.arange(self.nsn))
            j.append(idxc)
            v.append(dmu_dc)
            v, i, j = np.hstack(v), np.hstack(i), np.hstack(j)
            jacobian = scipy.sparse.coo_matrix((v, (i, j)), shape=(self.nsn, 3*self.nsn))
            return mu, jacobian
        return mu

    def plot(self, hessian, give_inv=False, save_dir=None):
        """
        Plot HD !

        Parameters
        -------
        hessian : numpy.ndarray
            Half of the hessian matrix given from minimizer or
            Inverse of half hessian matrix [if give_inv is True]
        give_inv : bool
            If the give hessian is already inversed.
        save_dir : None or str
            Saving directory of the plot.
        """
        #        from pycosmo import cosmo
        from astropy import cosmology
        import nacl.plotting.plots as plots
        from nacl.plotting.plots import hist
        #        cos = cosmo.Cosmo()
        cos = cosmology.FlatLambdaCDM(H0=70, Om=0.25)

        z, mu, dmu, var, inv_hessian = self.propagated_error(hessian, give_inv=give_inv)

        #        mu_th = cos.mu(z)
        mu_th = 5 * np.log10(np.array(cos.luminosity_distance(z))) + 25 # because pycosmo distances are in Mpc
        idx = z.argsort()

        plots.plot_HD(z, mu, dmu, mu_th, save_dir=save_dir)

        fig = pl.figure()
        chi2 = (mu - mu_th) / dmu
        ax = fig.add_subplot(111)
        hist(ax, chi2)
        ax.set_title(f'{(((mu - mu_th) / dmu) ** 2).sum()}')
        print('blue ', save_dir)
        if save_dir is not None:
            pl.savefig(save_dir + os.sep + f'distances_modulus_chi2.png')

        cov = var.copy()[idx].T[idx].T
        sig = np.sqrt(np.diag(cov))
        fig = pl.figure()
        ax0 = fig.add_subplot(121)
        m0 = ax0.matshow(cov, cmap=pl.matplotlib.cm.Blues)
        ax0.set_title('Distance covariance Matrix')
        pl.colorbar(m0)

        ax = fig.add_subplot(122)
        m = ax.matshow(cov / (sig.reshape(-1, 1) @ sig.reshape(-1, 1).T), cmap=pl.matplotlib.cm.Blues)
        ax.set_title('Distance correlation Matrix')
        pl.colorbar(m)
        if save_dir is not None:
            pl.savefig(save_dir + os.sep + f'distances_modulus_correlationMatrix.png')

    def propagated_error(self, hessian, give_inv=False):
        r"""
        From model fit, calculate distance modulus.

        The covariance matrix of the estimated parameters  (:math:`X_0`, :math:`X_1` and :math:`c`)
        can be obtained from the inverse of the hessian computed at the minimum.

        Let us note :math:`V^F = H^{-1}`, then the covariance matrix after marginalization on all the other
        parameters of the model is written~:


        .. math::
            V_{SN}= \begin{pmatrix}
            V^F_{X_0,X_0} & V^F_{X_0, X_1} & V^F_{X_0, c} \\
            V^F_{X_1, X_0} & V^F_{X_1, X_1} & V^F_{X_1, c} \\
            V^F_{c, X_0} & V^F_{c, X_1} & V^F_{c, c} \\
            \end{pmatrix}


        The covariance matrix of :math:`\mu` is deduced via the uncertainty propagation formula~:

        .. math::
            V_{\mu}^F(\mu) = J_{\mu} V_{SN} J_{\mu}^T

        where :math:`J_{\mu}} = (\frac{\partial }{partial X_0 },
        \frac{\partial }{partial X_1 }, \frac{\partial }{partial c})`
        is the matrix of derivatives of the distance modulus.

        Parameters
        -------
        hessian : numpy.array
            Half of the hessian matrix given from minimizer or
            Inverse of half hessian matrix [if give_inv is True]
        give_inv : bool
            If the give hessian is already inversed.

        Returns
        -------
        z : numpy.array
            SNe redshift.
        mu : numpy.array
            Distance modulus
        dmu : numpy.array
            Error on distance modulus from error on SNe parameters from fit.
        var : numpy.ndarray
            Covariance matrix.
        inv_hessian : numpy.ndarray
            Inverse of half of the hessian calculated at the minimum of the fit.
        """
        mu, jacobian = self(self.model.pars.free, jac=True)

        if give_inv:
            inv_hessian = hessian.copy()
        else:
            inv_hessian = np.linalg.inv(hessian)
        cov_matrix = inv_hessian[:3 * self.nsn, :3 * self.nsn]
        var = jacobian @ cov_matrix @ jacobian.T
        dmu = np.sqrt(np.diag(np.abs(var)))
        z = self.model.training_dataset.sne["z"]

        ret = [z, mu, dmu, var, inv_hessian]
        return ret



""" Implements a simple Spectrum class where: spec.x is the wavelength spec.y is
the flux spec.v is the variance

plus some methods deemed useful for SNe Ia (dereddening, deredshifting,
etc). Those methods are first implemented as functions working on arrays at the
namespace level so that they could be used more generally. They are then wrapped
into Spectrum class methods.

Also contains at the namespace level functions to read and create Spectrum
instances from files.

Note that for now, the only accepted file interface is the snfit ntuple format


Depends on croaks """


from . import interpolation, robuststat, constants
from ..util import saltpath, io

# temporary change, until pyfits is discontinued
try:
    import astropy.io.fits as pyfits
except:
    import pyfits
import numpy as np
import scipy
import scipy.signal
import copy
import pdb
import re
import sys
import os
import os.path as op
import glob
import math

def load_snf_std_fits(filename, verbose = False):
    """ Reads SNfactory reference
    spectra in fits format, as found in
    CVS_SNf/Tasks/Calibration/Reftables/Refflux/fits

    implementation stolen from
    CVS_SNf/Tasks/Calibration/Reftables/Refflux/tools/plot_refflux.py """

    t = pyfits.open(filename)
    tbdata = t[1].data
    
    # Get lambda and select proper range
    if 'WAVELENGTH' in t[1].columns.names:
        lbdaCol = 'WAVELENGTH'
    else:
        lbdaCol = 'LAMBDA'
    lbda = tbdata.field(lbdaCol).copy()
    lbdaUnits = t[1].columns.units[t[1].columns.names.index(lbdaCol) ]

    if verbose:
        print(" Wavelength from %s column: %.2f,%.2f '%s'" %  (lbdaCol, lbda[0], lbda[-1], lbdaUnits))
    
    flux = tbdata.field('FLUX').copy()
    fluxUnits = t[1].columns.units[t[1].columns.names.index('FLUX') ]
    if verbose:
        print(" Flux from FLUX column in '%s'" % fluxUnits)

    t.close()

    return Spectrum(lbda, flux)
    

def load_snf_fits(filename):
    """ Reads a spectrum object from an SNf fits file.
    Puts all the header information in a dictionnary s.keys

    Note that it should also work for all fits spectra, but in that case we
    can't be sure that the second extension is actually the variance.

    If no second extension is found, it doesn't crash, it simply doesn't read
    it.  """
    spec_fits = pyfits.open(filename)

    #- read the fits data +
    name = spec_fits[0].header.get('FILENAME')
    npts = spec_fits[0].header['NAXIS1']
    step = (spec_fits[0].header['CDELT1'])
    start = (spec_fits[0].header['CRVAL1'] - (spec_fits[0].header.get('CRPIX1', 1) -1) * step)
    end = start + (npts - 1) * step
    x = np.linspace(start, end,npts)
    y = spec_fits[0].data.astype('d')

    #- reads the variance, assumed to be
    #in the second fits hdu
    if len(spec_fits) == 2:
        v = spec_fits[1].data.astype('d')
    else: v = None

    s = Spectrum(x, y, v=v)
    s.keys = {}
    for key, value in spec_fits[0].header.items():
        s.keys[key] = value
        
    return s


def load_ntuple(filename, wavelength_key="WAVE", flux_key="FLUX", error_key="ERR"):
    """ Reads a spectrum from an snfit ntuple format file
    Keywords are: WAVE FLUX ERR VALID

    Note that the keys of the ntuple are propagates in s.keys for convenience
    """
    nt_s = io.NTuple.fromtxt(filename)

    if error_key in nt_s.dtype.names:
        s = Spectrum(nt_s[wavelength_key], nt_s[flux_key], v=nt_s[error_key] ** 2)
    else:
        s = Spectrum(nt_s[wavelength_key], nt_s[flux_key])
        
    s.keys = nt_s.keys

    return s


def read_ascii(filename):
    """
    Read ascii files and returns a list of lines.
    Empty lines and lines beginning with # are not returned.
    All comments starting with # in a line are dropped.
    All lines are stripped.
    """
    fh = open(filename, "r")
    line_list = []
    for line in fh.readlines():
        line = (re.sub("#.*$", "", line)).strip()        
        if len(line) == 0:
            continue
        line_list.append(line)
    return line_list


def load_ascii(filename):
    """
    Useful to read a reference spectrum, as pointed at by
    MagSys

    Reads a spectrum object from an ascii file, assuming the format:
    wavelength   flux   variance
    
    Variance being optional, and # delimitating comments.
    Empty lines ignored

    Since I am not sure if the third column is really variance, I skip it
    And since there can be more than 3 columns I just keep the 2 first ones
    """

    l_line = read_ascii(filename)

    x = []
    y = []
#    if len(l_line[0].split()) == 3:
#        v = []
#        for line in l_line:
#            w, f, variance = line.split()
#            x.append(float(w))
#            y.append(float(f))
#            v.append(float(variance))
#        return Spectrum(np.array(x), np.array(y), v=np.array(v))
#    else:
#        for line in l_line:
#            w, f = line.split()
#            x.append(float(w))
#            y.append(float(f))
#        return Spectrum(np.array(x), np.array(y))
    for line in l_line:
        if line[0] in ['@', '#']:
            continue
        line = line.split()
        x.append(float(line[0]))
        y.append(float(line[1]))
    return Spectrum(np.array(x), np.array(y))


def load_sn_template(phase, filename="snflux_1a_hsiao.dat"):
    """
    Reads a spectrum object from an ascii file, assuming the format:
    date  wavelength   flux
    
    It will return the spectrum corresponding to the phase asked for. If the phase doesn't exist, it will crash.


    If filename is a path, it will use the path. If is it just a basename
    it will look for it in 
    saltpath.sn_templates

    By default loads Hsiao template
    """
    if len(os.path.dirname(filename)) == 0:
        filename = os.path.join(saltpath.sn_template_path, filename)
        
    x = []
    y = []    
    for line in read_ascii(filename):
        p, w, f = line.split()
        if float(p) == phase:
            x.append(float(w))
            y.append(float(f))
    
    return Spectrum(np.array(x), np.array(y))


def load_galaxy_template(galaxy_type="S0", age=None, path=saltpath.galaxy_template_path, no_clip=False):
        """
        galaxy_type is 
           E
           S0
           Sa
           Sb
           Sc

        Age is from 2gy to 20gy by integer increments
        If age is provided will load a Pegase galaxy, otherwise will load a
        Kinney (check spelling) galaxy

        If no path towards the template directory given, will assume that they
        live in $SALTPATH/galaxy_templates/
        
        If no_clip is True returns the full template, otherwise clips it between
        2000 and 10000 AA
        """

        if age is None:
            filename = "galaxy_template_%s.dat" % (galaxy_type)
        else:
            filename = "gal_peg_%s_%s.dat" % (galaxy_type, age)

        s = load_ascii(os.path.join(path, filename))
        if not no_clip:
            s = s.clip(w_min=2000., w_max=10000.)
        return Spectrum(s.x, s.y)


def load_pickles(name=None, path=saltpath.stellar_libs_path, no_clip=False):
    """
    Load the Pickles library and return a dictionary of all the
    spectra, indexed by the stellar type names.
    
    We work directly from the original ASCII files, obtained from CDS.
    We read only the first two columns, which correspond to characters
    1-7 and 9-17 resp.
    
    Wavelengths are in Angstroems. 
    Fluxes are in [Energy/s/A/cm^2], normalized as: f(lambda)/f(5556 A)
    
    The third column (standard deviation, when several spectra are
    combined in ignored, as are all subsequent columns).
    
    """
    ret = []
    path = op.join(saltpath.stellar_libs_path, 'pickles')
    if name is not None:
        filenames = ['%s.dat' % name]
    else:
        filenames = glob.glob(op.join(path, '*.dat'))
    for fn in filenames:
        f = open(fn)
        wl, flx = [], []
        for l in f:
            wl.append(float(l[0:7]))
            flx.append(float(l[8:17]))
        s = Spectrum(np.array(wl), np.array(flx))
        stellar_type_name = op.basename(fn).split('.dat')[0]
        s.keys['name'] = stellar_type_name
        ret.append(s)
        f.close()
    return ret


#- All the variance stuff will go in Variance_Spectrum
#- with its own factories, interpolators etc.
class Spectrum:
    def __init__(self, x, y, v=None, keys={}):
        """
        This is a very simple spectrum class, without safeguards of any kinds.
        
        self.x is the wavelength, assumed to be in AA
        self.y is the flux, assumed to be in ergs per sec per AA per cm^2
        self.v is the variance corresponding to self.x. Can be None.

        self.flux() is a interpolation Func1D that interpolates the flux
        self.photon() is a interpolation Func1D that interpolates the number
        of photons

        self.keys is a place holder for the keys of a spectrum nacl.util.io ntuple.
        It is initialized to None.

        WARNING: almost no method propagates the variance for now
        """

        self.x = x
        self.y = y
        self.v = v
        self.keys = keys

        #- The names here don't matter, they are lost in creating the Func1D
        self.flux = interpolation.Func1D(x, y)
        
        self.photon = interpolation.Func1D(x, flambda_to_fclambda(x, y))
        self.p = self.photon(self.x)

    
    def totxt(self, filename=None, comment=None, light=False, extra_light=False, wavelength_scale=1.):
        """
        Prints the spectrum to a file in NTuple format.
        If no filename is provided, prints to stdout.
        
        light: only prints wavelength flux and variance
        extra_light: only prints wavelength and flux

        wavelength_scale : multiplicator to wavelength, for example to get from AA to nm

        In order to comply with SALT2 idiosyncrasies, the keywords are:
        Keywords are:
        WAVE
        FLUX
        ERR
        VALID
        """
        
        if filename is None:
            fh = sys.stdout
        else:
            fh = open(filename, "w")

        if extra_light:
            light = True
            
        if light:
            if self.v is None or extra_light:
                for wave, flux in zip(self.x * wavelength_scale, self.y):
                    print("%-f    %-e" % (wave, flux), file=fh)
                return
            else:
                for wave, flux, v in zip(self.x * wavelength_scale, self.y, self.v):
                    print("%-f    %-e    %-e" % (wave, flux, v), file=fh)
                return


        if comment is not None:
            print("@generation_comment: %s" % (comment), file=fh)
        
        for key, value in self.keys.items():
            print("@%s : %s" % (key, str(value)), file=fh)
        
        if self.v is None:
            print("#WAVE:", file=fh)
            print("#FLUX:", file=fh)
            print("#VALID:", file=fh)
            print("#end", file=fh)
            
            for wave, flux in zip(self.x * wavelength_scale, self.y):
                print("%-f    %-e    1" % (wave, flux), file=fh)

        else:
            print("#WAVE:", file=fh)
            print("#FLUX:", file=fh)
            print("#ERR:", file=fh)
            print("#VALID:", file=fh)
            print("#end", file=fh)

            for wave, flux, variance in zip(self.x * wavelength_scale, self.y, self.v):
                err = np.sqrt(variance)
                print("%-f    %-e    %-e    1" % (wave, flux, err), file=fh)

    def clip(self, w_min=None, w_max=None):
        """
        returns a clipped spectrum with wavelength between the two bounds
        given in options.
        
        """
        
        if w_min is None:
            w_min = np.min(self.x)
        if w_max is None:
            w_max = np.max(self.x)

        i_ok = (self.x < w_max) & (self.x > w_min)
        if self.v is not None:
            v = self.v[i_ok]
        else:
            v = None
        return Spectrum(self.x[i_ok], self.y[i_ok], v=v, keys=self.keys)

    def noise_up(self, S_to_N = 10):
        """
        Returns a spectrum noised up to a S/N of S_to_N w.r.t the
        initial spectrum.
        This does *not* take into account the original noise of the spectrum.

        WARNING: assumes that the spectrum is flux calibrated
        """
        y_photon = flambda_to_fclambda(self.x, self.y)
        if S_to_N <= 0.:
            y_noise = y_photon
            v = copy.copy(self.v)
        else:
            mean_y_photon = np.mean(y_photon)
            y_noise = y_photon/mean_y_photon * S_to_N**2
            v = copy.copy(y_noise)
    
            y_noise = scipy.random.poisson(v)
    
            y_noise = y_noise *  mean_y_photon / S_to_N**2
            v = v * (mean_y_photon / S_to_N**2)**2
    
        y_noise = fclambda_to_flambda(self.x, y_noise)
        #- Twice for the variance
        v = fclambda_to_flambda(self.x, v)
        v = fclambda_to_flambda(self.x, v)
    
        return Spectrum(self.x, y_noise, v=v, keys=self.keys)



    def rescale_to_mag(self, mag_target, instrument="SWOPE", magsys="VEGA", current_filter="g"):
        """
        Returns a spectrum rescaled in ergs/cm^s/s/AA so that its magnitude in
        band [band] is equal to mag_target

        instrument and  magsys  can be either strings (and the saunerie
        equivalents will be loaded) or directly the saunerie equivalents.

        if current_filter is a Transmission instance, no need to provide
        instrument
        """

        if type(instrument) == str:
            instrument = instruments.InstrumentModel(instrument)
        if type(magsys) == str:
            magsys = instruments.MagSys(magsys)
        if type(current_filter) == str:
            current_filter = instrument.EffectiveFilterByBand(current_filter)


        mag_orig = magsys.mag(current_filter, self)[0][0]
        alpha = 10 ** (0.4 * (mag_orig - mag_target))
        return  spectrum.Spectrum(self.x, self.y * alpha, v=self.v, keys=self.keys)

    
                
    def deredden(self, ebmv=0., Rv=3.1):
        """
        deredden implementation as Spectrum method
        """        

        if self.v is not None:
            new_y, ccm_factor = deredden(self.x, self.y, ebmv=ebmv, Rv=Rv, return_ccm_factor=True)
            new_v = self.v * ccm_factor ** 2
            return Spectrum(self.x, new_y, v=new_v, keys=self.keys)
        else:
            new_y = deredden(self.x, self.y, ebmv=ebmv, Rv=Rv)
            return Spectrum(self.x, new_y, keys=self.keys)

    def redden(self, ebmv=0., Rv=3.1):
        """
        deredden implementation as Spectrum method

        Doesn't do anything to the variance
        """

        new_y = redden(self.x, self.y, ebmv=ebmv, Rv=Rv)
        return Spectrum(self.x, new_y, v=new_v, keys=self.keys)

    def rebin_simple(self, n):
        """
        Rebins the spectrum by averaging over n pixels.

        Also rebins the variance in the same way
        """
        if n <= 1:
            return Spectrum(self.x, self.y, v=self.v, keys=self.keys)
        
        x, y = rebin_simple(self.x, self.y, n)
        if self.v is not None:
            x, v = rebin_simple(self.x, self.v, n)
        else:
            v = None
        return Spectrum(x, y, v=v,  keys=self.keys)

    def SG_filter(self, n_bin, degree=3):
        """
        Uses the Savitzky_Golay implementation here to return a smoothed spectrum

        FIXME: need to implement the correct handling of variance
        """
        sg = Savitzky_Golay(n_bin, pol_degree=degree)

        if self.v is not None:
            print("WARNING: for now SG_filter drops the variance")
        
        return Spectrum(self.x, sg(self.y), keys=self.keys)

    def median_filter(self, kernel_size=30):
        """
        Uses scipy.signal.medfilter to median filter the spectrum
        """
        if self.v is not None:
            print("WARNING: for now  drops the variance")
        
        return Spectrum(self.x, scipy.signal.medfilt(self.y, kernel_size=kernel_size), keys=self.keys)        

    def slide_win(self,width=5, func=np.mean):
        """
        Applies np.mean on a sliding window of width 5 by default.
        func provides a differenc function to apply
        width changes the width on which the function is applied
        """
        if self.v is not None:
            print("Warning: for now slide_win drops the variance")

        return Spectrum(self.x, robuststat.slide_win(self.y, func=func, width=width), keys=self.keys)
            
    
    def deredshift_wavelength(self, z=0.):
        """
        Deredshifts only the wavelength axis (no change on flux)
        If z < 0 applies a blueshift
        """
        #- doppler_shift applies the redshift, therefore deredshifting has to go the other direction
        #- Thus the - sign
        new_x = doppler_shift(self.x, z=-z)
        return Spectrum(new_x, self.y, v=self.v, keys=self.keys)

    def redshift_wavelength(self, z=0.):
        """
        Deredshifts only the wavelength axis (no change on flux)
        If z < 0 applies a blueshift
        """
        #- doppler_shift applies the redshift, therefore deredshifting has to go the other direction
        #- Thus the - sign
        new_x = doppler_shift(self.x, z=z)
        return Spectrum(new_x, self.y, v=self.v, keys=self.keys)

    def to_rest_frame(self, z_helio=None, z_cmb=None, idr_restframe_scale=False):
        """
        Deredshifts the wavelength axis by z_cmb (or z_helio if z_cmb is not provided)
        and does *not* blueshift it back to z_ref, unless idr_restframe_scale is set to True
        
        """

        new_x = wavelength_to_rest_frame(self.x, z_helio=z_helio, z_cmb=z_cmb)
        new_y = flux_to_rest_frame(self.y, z_helio=z_helio, z_cmb=z_cmb, idr_restframe_scale=idr_restframe_scale)
            
        return Spectrum(new_x, new_y, v=self.v, keys=self.keys)

    def to_observer_frame(self, z_helio=None, z_cmb=None, idr_restframe_scale=False):
        """
        Redshifts the wavelength and the flux by z_cmb and z_helio.
        If only one is given, only uses it.        
        """
        new_x = wavelength_to_observer_frame(self.x, z_helio=z_helio, z_cmb=z_cmb)
        new_y = flux_to_observer_frame(self.y, z_helio=z_helio, z_cmb=z_cmb, idr_restframe_scale=idr_restframe_scale)

        return Spectrum(new_x, new_y, v=self.v, keys=self.keys)

    
    #- snfit IntegFlux considers the SN spectrum to be restframe but the filter to be
    #- observer frame. The z provided is meant to redshift the SN to there.
    #- If I have a SN spectrum observer frame, providing z=0 will yield 
    #- the observer frame magnitude of the SN at z.
    #-    number of photons = energy/(h*nu) = energy * wl/(2*pi*hbar*c)
    #-    2*pi* hbar*c = 2* pi * 197 eV nm = 6.28*197 * 1.60e-12 * 10 ergs angst
    #-    = 1.97946e-08
    #-    = 1/5.006909561e07
    def IntegFlux(self, effective_filter, z=0., wavelength_integration_step=5):
        """
        effective_filter is the filter transmission provided as Filter instance (defined in instruments.py derives from Func1D)
    
        This integration function follows snfit conventions:
    
        The spectrum flux is assumed to be in energy, i.e. in ergs/s/cm^2
    
        The filter transmission is assumed to be in pseudo_ADU/photon
    
        The redshift is the redshift between the observer frame and the spectrum
        frame. If the spectrum is observer frame then z=0 means an observer frame integral. 
    
        If the spectrum is restframe, z=0 means that restframe and observer frame are the 
        same. 
    
        The case z !=0 needs special attention. The spectrum is redshifted to redshift z
        and the filters remain in observer frame. The filters are thus blueshifted w.r.t 
        the spectrum. It is the typical case of observing a supernova at redshift z in 
        an observer frame filter F knowing its restframe spectrum S.
        A special warning: the spectrum is only redshifted as far as wavelength is concerned.
        The flux dillution due to the luminosity distance of the supernova is 
        NOT taken into account.
    
        wavelength_integration_step is kept as an option for convenience.
        To follow snfit it can be set looking for WAVELENGTH_INTEGRATION_STEP  
        in fitmodel.card
        snfit default, as this function default is an integration step of 5 AA.
    
    
        Note that in ALL CASES the spectrum wavelength axis and the filter wavelength
        axis are in the SAME FRAME.
        """
        
        spectrum_shifted = self.redshift_wavelength(z=z)
        
        nphot_l = spectrum_shifted.photon * effective_filter
        
        # Calculation of  the number of integration steps:
        n_steps = int((effective_filter.x_max - effective_filter.x_min) / wavelength_integration_step)
        
        # the wavelength integration step size is taken into account 
        # by Func1D integrator. 
        n_photon = nphot_l.integrate(effective_filter.x_min, effective_filter.x_max, n_steps)

        # This is to account for the flux dilution in a delta_lambda step, and to account for the fact
        # that snfit IntegFlux returns the flux in the blueshifted filter, not in restframe filters.
        n_photon /= (1. + z)**2
        
        # NB: hc for snfit and for saunerie are slightly different:    
        n_photon *= 5.006909561e7 * (constants.CLIGHT_A_s * constants.H_e_s)
        return n_photon

    
            
        

    
    
#- Spectrum manipulation, implemented as array operations    
def rebin_simple(x, y, n):
    """ Bins up the spectrum by averaging the values of every n
    pixels.

    """
    remain = len(x) % n
    if remain > 0:
        x = x[:-remain].reshape(-1, n)
        y = y[:-remain].reshape(-1, n)
    else:
        x = x.reshape(-1, n)
        y = y.reshape(-1, n)

    n = float(n)
    x = np.nansum(x, axis=1) / n
    y = np.nansum(y, axis=1) / n
    
    return x, y


def deredden(x, y, ebmv=0., Rv=3.1, return_ccm_factor=False):
    """
    Deredden flux (y) using Cardelli,Clayton,Mathis law 
    
    x is the wavelength
    y is the flux

    ebmv is E(B-V)
    Rv is Rv, default value is 3.1

    """    
    ccm_factor = 1. / ccm(x, ebmv, Rv=Rv) 
    y_tmp = y * ccm_factor
    if return_ccm_factor:
        return y_tmp, ccm_factor
    else:
        return y_tmp
    


def redden(x, y, ebmv=0., Rv=3.1):
    """
    Redden flux (y) using Cardelli,Clayton,Mathis law 
    
    x is the wavelength
    y is the flux
    ebmv is E(B-V)
    Rv is Rv, default value is 3.1

    """    
    #- Only difference with deredden is ccm_factor == ccm instead of 1./ccm
    ccm_factor = ccm(x, ebmv, Rv=Rv) 
    y_tmp = y * ccm_factor
    return y_tmp


def doppler_shift(x, z=0.):
    """
    Shifts a wavelength array by 
       * (1.+z) if  z > 0.
       / (1.-z) if z < 0.
    """
    if (z == 0.) or (z is None):
        return x
    if z > 0.:
        return x * (1.0 + z)
    if z < 0.:
        return x / (1.0 - z)

    
def wavelength_to_rest_frame(x, z_helio=None, z_cmb=None):
    """
    Deredshifts the wavelengths using z_cmb is it is provided, z_helio otherwise.
    If both are none, returns the wavelength unchanged.
    This is supposed to be used together with flux_to_rest_frame
    """
    if z_cmb is None:
        if z_helio is not None:
            new_x = doppler_shift(x, - z_helio)
        else:
            new_x = x
    else:
        new_x = doppler_shift(x, - z_cmb)

    return new_x


def flux_to_rest_frame(y, z_helio=None, z_cmb=None, idr_restframe_scale=False):
    """
    Deredshifts the flux accounting for the luminosity distance dillution as calculated in
    luminosity_distance

    If only one of z_helio or z_cmb are provided, both are assumed to be the same
    The restframe is considered to be at 10pc, where 
    absolute magnitudes are defined.
    """
    if (z_helio is None) and (z_cmb is None):
        return y
    if z_helio is not None:
        if z_cmb is None:
            z_cmb = z_helio

    if z_cmb is not None:
        if z_helio is None:
            z_helio = z_cmb

    #- Reference redshift is restframe at 10pc
    if idr_restframe_scale:
        z_ref = 0.05    
        d_r = luminosity_distance(z_helio=z_ref, z_cmb=z_ref)
        factor = 1.e15
    else:
        z_ref = 0.    
        #- 10 pc = 10.e-6 Mpc
        d_r = 10.e-6  
        factor = 1.

    #- Luminosity distance
    d_z = luminosity_distance(z_helio=z_helio, z_cmb=z_cmb)

    factor *= (1.0 + z_helio) / (1.0 + z_ref) * d_z * d_z / (d_r * d_r)
    new_y = y * factor

    return new_y


def wavelength_to_observer_frame(x, z_helio=None, z_cmb=None):
    """
    Redshifts the wavelengths using z_cmb is it is provided, z_helio otherwise.
    If both are none, returns the wavelength unchanged.
    This is supposed to be used together with flux_to_rest_frame
    """
    if z_cmb is None:
        if z_helio is not None:
            new_x = doppler_shift(x, z_helio)
        else:
            new_x = x
    else:
        new_x = doppler_shift(x, z_cmb)

    return new_x


def flux_to_observer_frame(y, z_helio=None, z_cmb=None, idr_restframe_scale=False):
    """
    Redshifts the flux accounting for the luminosity distance dillution as calculated in
    luminosity_distance.
    It assumes that the SN flux was given for a supernova at z_ref. 
    By default, it assumes that the original flux is given at 10pc

    If only one of z_helio or z_cmb are provided, both are assumed to be the same
    If no z_ref is provided, the restframe will be considered to be at 10pc, where 
    absolute magnitudes are defined.
    """

    if (z_helio is None) and (z_cmb is None):
        return y

    if z_helio is not None:
        if z_cmb is None:
            z_cmb = z_helio

    if z_cmb is not None:
        if z_helio is None:
            z_helio = z_cmb

    #- Reference redshift is restframe at 10pc
    #- Reference redshift is restframe at 10pc
    if idr_restframe_scale:
        z_ref = 0.05    
        d_r = luminosity_distance(z_helio=z_ref, z_cmb=z_ref)
        factor = 1.e15
    else:
        z_ref = 0.   
        #- 10 pc = 10.e-6 Mpc 
        d_r = 10.e-6 
        factor = 1

    #- Luminosity distance
    d_z = luminosity_distance(z_helio=z_helio, z_cmb=z_cmb)

    factor *= (1.0 + z_helio) / (1.0 + z_ref) * d_z * d_z / (d_r * d_r)

    new_y = y / factor
    return new_y
    
#--- Tools
def locate(w, w_min=None, w_max=None, method="range"):
    """
    Finds index for which w is between w_min and w_max.
    If either w_min or w_max are None, they are taken as the extremes of the array.

    If method is "range" returns all the indexes.
    If method is min, max or mean, returns respectively the min, the max or the median
    of all the indexes
    """
    if w_min is None:
        w_min = min(w)
    if w_max is None:
        w_max = max(w)

    i_ok = np.where((w < w_max) & (w >w_min))
    i_ok = i_ok[0]
    
    if method == "min":
        return np.min(i_ok)
    elif method == "max":
        return np.max(i_ok)
    elif method == "median":
        return int(np.median(i_ok))
    elif method == "range":
        return i_ok
    else:
        return list(range(len(w) - 1))
        print("Method not implemented, returning the full range")

#- Astronomy

#- unit transformation
CLIGHT_A_s = 2.99792458e18         # [A/s]
CLIGHT_m_s = 2.99792458e8         # [A/s]
CLIGHT_km_s = 2.99792458e5         # [A/s]
HPLANCK = 6.62606896e-27        # [erg s]


def lambda_to_nu(x):
    """
    Transforms wavelength in AA in Hertz
    """    
    return CLIGHT_A_s / x


def nu_to_lambda(x):
    """
    Transforms Hertz in AA 
    """
    return CLIGHT_A_s / x


def flambda_to_fclambda(x, y):
    """
    Transforms a flux in erg/s/cm^2/AA in flux in photons/s/cm^2
    x in AA

    n_l*d_l = f_l*d_l / (h*nu)
    n_l*d_l = f_l*d_l * l/h*c
    """    
    return y * x / (CLIGHT_A_s * HPLANCK)


def fclambda_to_flambda(x, y):
    """
    Transforms a flux in erg/s/cm^2/AA in flux in erg/s/cm^2/photons
    x in AA

    n_l*d_l = f_l*d_l * l/(h*c)
    f_l*dl = h*c/l * n_l*d_l
    """
    
    return y * CLIGHT_A_s * HPLANCK / x


def fnu_to_flambda(frequency, flux):
    """
    Returns flambda that corresponds to nu_to_lambda(frequency) wavelength array
    
    f_l = f_nu c/lambda^2
    """
    
    wavelength = nu_to_lambda(frequency)
    return flux * CLIGHT_A_s / wavelength ** 2
    

#- Flux to magnitude:
def magnitude(fx, dfx, m0=48.59, zp=0.):
    """
    Calculates a magnitude from a flux and its error
    """
    if fx <= 0 or np.isnan(fx):  # Nothing we can do about negative flux!
        return np.nan, np.nan
    else:
        m = -2.5 * np.log10(fx) + zp - m0

    if dfx > 0:
        dm = 2.5 / np.log(10) / fx * dfx
    else:
        dm = 0

    return m, dm



#- CCM reddening law 
#- Stolen from Prospect/Util.py and updated from O'Donnell
def ccm(wavelength, ebmv, Rv=3.1):
    """
    Given a wavelength in angstroms, calculate the
    extinction A(w) according to the CCM model.
    c.f. Cardelli, Clayton, Mathis 1989: ApJ 345:245
    with optical/NIR part updated from
    O'Donnell (1994ApJ...422..158O).

    :Parameters:
        `wavelength` : np.array
            Array of wavelength values at which you which to
            find warp values
         `ebmv` : float
         `Rv` : float, default is 3.1


    :return: np.array
    """
    # change variables from angstroms to microns
    ang_to_micron = 1.0e-4
    # cardelli's parameter x = 1/(wavelength in microns)
    x = 1.0 / (wavelength * ang_to_micron)

    # version to deal with arrays
    ex1 = (x < 1.1) * __ccm_ir(x, Rv)
#    ex2 = (1.1 <= x  ) * (x < 3.3) * __ccm_optical(x, Rv)
    ex2 = (1.1 <= x) * (x < 3.3) * __odonnell_optical(x, Rv)
    ex3 = (3.3 <= x) * (x < 8.0) * __ccm_uv1(x, Rv)
    ex4 = (x >= 8.0) * __ccm_uv2(x, Rv)
    extinction = ex1 + ex2 + ex3 + ex4

    # get flux multiplicative factor from the extinction
    return 10.0 ** (-0.4 * extinction * ebmv)


def __ccm_ir(x, rv):
    # for x < 1.1
    a0 = 0.574
    b0 = -0.527
    c0 = rv * a0 + b0
    return c0 * (x ** 1.61)


def __ccm_optical(x, rv):
    # for 1.1 < x < 3.3
    y = x - 1.82
    # a coeffs
    a0 = 1.0
    a1 = 0.17699
    a2 = -0.50447
    a3 = -0.02427
    a4 = 0.72085
    a5 = 0.01979
    a6 = -0.77530
    a7 = 0.32999
    # b coeffs
    b0 = 0.0
    b1 = 1.41338
    b2 = 2.28305
    b3 = 1.07233
    b4 = -5.38434
    b5 = -0.62251
    b6 = 5.30260
    b7 = -2.09002
    # total coeffs
    c0 = rv * a0 + b0
    c1 = rv * a1 + b1
    c2 = rv * a2 + b2
    c3 = rv * a3 + b3
    c4 = rv * a4 + b4
    c5 = rv * a5 + b5
    c6 = rv * a6 + b6
    c7 = rv * a7 + b7
    # finally result
    return (c0
             + c1 * y
             + c2 * (y ** 2)
             + c3 * (y ** 3)
             + c4 * (y ** 4)
             + c5 * (y ** 5)
             + c6 * (y ** 6)
             + c7 * (y ** 7))


def __odonnell_optical(x, rv):
    """A/Av in Optical/Near IR: 1.1-3.3 micron^-1"""
    y = x - 1.82
    pa = [-0.505, +1.647, -0.827, -1.718, +1.137, +0.701, -0.609, +0.104, 1]
    pb = [+3.347, -10.805, +5.491, +11.102, -7.985, -3.989, +2.908, +1.952, 0]
    a = np.polyval(pa, y)
    b = np.polyval(pb, y)
    return a * rv + b


def __ccm_uv1(x, rv):
    # for 3.3 < x < 8.0
    y = x - 5.90
    fa = (x >= 5.90) * (-0.04473 * (y ** 2) - 0.009779 * (y ** 3))
    fb = (x >= 5.90) * (0.21300 * (y ** 2) + 0.120700 * (y ** 3))
    aa = 1.752 - 0.316 * x - 0.104 / ((x - 4.67) ** 2 + 0.341) + fa
    bb = -3.09 + 1.825 * x + 1.206 / ((x - 4.62) ** 2 + 0.263) + fb
    return rv * aa + bb


def __ccm_uv2(x, rv):
    # for x>8.0
    y = x - 8.0
    # a coeffs
    a0 = -1.073
    a1 = -0.628
    a2 = 0.137
    a3 = -0.070
    # b coeffs
    b0 = 13.670
    b1 = 4.257
    b2 = -0.420
    b3 = 0.374
    # total coeffs
    c0 = rv * a0 + b0
    c1 = rv * a1 + b1
    c2 = rv * a2 + b2
    c3 = rv * a3 + b3
    return (c0
             + c1 * y
             + c2 * (y ** 2)
             + c3 * (y ** 3))


#- Luminosity distance
def luminosity_distance(z_helio, z_cmb, hubble=70.0, omega_m=0.28):
    """
    Returns luminosity distance using equation (16) of Kantowski
    & Thomas (2001), ApJ, 561, 491 with the CMB redshift used in the
    integral but heliocentric redshift used everywhere else.

    Luminosity distance returned in Mpc
    """
    import math

    hubble = 70.0
    omega_m = 0.28
    sol = 299792.458
    front = 2.0 * (1.0 + z_helio) * sol / hubble / math.pow(omega_m, 1.0 / 3.0)  # note z_helio here
    a = 1.0 / 6.0
    b = 2.0 / 3.0
    c = 7.0 / 6.0
    omom = 1.0 - omega_m
    arg = 1.0 + omega_m * z_cmb * (3.0 + z_cmb * (3.0 + z_cmb))  # note z_cmb here
    return front * (scipy.special.hyp2f1(a, b, c, omom) - math.pow(arg, - a) * scipy.special.hyp2f1(a, b, c, omom / arg))


#- Stolen from SJB EMPCA:
class Savitzky_Golay(object):
    """
    Utility class for performing Savitzky Golay smoothing
    
    Code adapted from http://public.procoders.net/sg_filter/sg_filter.py
    """
    def __init__(self, width, pol_degree=3, diff_order=0):
        self._width = width
        self._pol_degree = pol_degree
        self._diff_order = diff_order
        self._coeff = self._calc_coeff(width//2, pol_degree, diff_order) 

    def _calc_coeff(self, num_points, pol_degree, diff_order=0):
    
        """
        Calculates filter coefficients for symmetric savitzky-golay filter.
        see: http://www.nrbook.com/a/bookcpdf/c14-8.pdf
    
        num_points   means that 2*num_points+1 values contribute to the
                     smoother.
    
        pol_degree   is degree of fitting polynomial
    
        diff_order   is degree of implicit differentiation.
                     0 means that filter results in smoothing of function
                     1 means that filter results in smoothing the first 
                                                 derivative of function.
                     and so on ...
        """
    
        # setup interpolation matrix
        # ... you might use other interpolation points
        # and maybe other functions than monomials ....
    
        x = np.arange(-num_points, num_points+1, dtype=int)
        monom = lambda x, deg : math.pow(x, deg)
    
        A = np.zeros((2*num_points+1, pol_degree+1), float)
        for i in range(2*num_points+1):
            for j in range(pol_degree+1):
                A[i,j] = monom(x[i], j)
            
        # calculate diff_order-th row of inv(A^T A)
        ATA = np.dot(A.transpose(), A)
        rhs = np.zeros((pol_degree+1,), float)
        rhs[diff_order] = (-1)**diff_order
        wvec = np.linalg.solve(ATA, rhs)
    
        # calculate filter-coefficients
        coeff = np.dot(A, wvec)
    
        return coeff
    
    def __call__(self, signal):
        """
        Applies Savitsky-Golay filtering
        """
        n = np.size(self._coeff-1)/2
        res = np.convolve(signal, self._coeff)
        return res[n:-n]

#- SN template class
class SnTemplate(object):
    @classmethod
    def load_ascii(self, filename, names=["phase", "wavelength", "flux"], **keyargs):
        """
        Uses np.genfromtxt to load

        usecols=(1,3,5) to select which columns to use
        names = ["a", "flux", "wavelength", "b", "phase","d"]
        or
        names = ["flux", "wavelength", "phase"] 

        By default, assumes that phase comes first, then wavelength then flux
        """

        nt = np.genfromtxt(filename, names=names, **keyargs)
        nt = nt.view(io.NTuple)

        return SnTemplate(nt["phase"], nt["wavelength"], nt["flux"])
    
    def __init__(self, phase, wavelength, flux):
        """
        Assumes that the phase is in days
                     wavelength is in AA
                     flux is in energy, and ergs/s/cm^2/AA even. But as long as you don't use the photon conversion, you are probably fine 
        """
        nt = np.rec.fromarrays([phase, wavelength,flux], names=["phase", "wavelength", "flux"])
        self.nt = nt.view(io.NTuple)
        self.dp = croaks.DataProxy(self.nt, p="phase", w="wavelength", f="flux")
        self.dp.make_index("p", intmap=True)

        

    def __call__(self, phase, method="nearest"):
        """
        Returns the Spectrum instance corresponding to the phase selected by the method
        
        nearest: the phase nearest to the phase asked for
        key "SnTemplate.phase" contains the phase
        """

        p_index = croaks.find_closest(self.dp.p_set, phase)
        i_ok = self.dp.p_index == p_index
        
        s = Spectrum(self.dp.w[i_ok], self.dp.f[i_ok])
        s.keys["SnTemplate.phase"] = self.dp.p_set[p_index]

        return s
        



import numpy as np
import nacl.instruments
import os 
from nacl.simulations.reindex import reindex
from nacl.dataset import TrainingDataset
import glob


class Generator(object):
    r"""
    Data generator.
    Generate K21 simulation (Kenworthy et al. 2021)

    We start again from the objects composing each of these training batches~: we take their redshifts,
    their SALT2 parameters (:math:`X_0`, :math:`X_1`, :math:`c`, :math:`t_{max}`) as well as the
    photometric and spectroscopic observation dates.


    Attributes
    ----------
    filespath : str
        Path to the repository containing the light curve et spectra files.
    filterpath : str
        Path to the filters files repository in nacl.dev/data/
    Sne ; numpy.array
        SNe names
    lc : numpy.rec.array
        Light curves data with data_type as type.
    sp : numpy.rec.array
        Spectral data, with data_type as type.
    snInfo : numpy.rec.array
        SNe information (:math:`z`, :math:`tmax`, :math:`x1`, :math:`x0`, :math:`c`)
    data_type : list
        NaCl data type.
    snInfo_type : list
        Type of information of SNe.

    sigma_lc : numpy.array
        Dispersion of photometric data
    sigma_sp : numpy.array
        Dispersion of spectral data.
    trainingDataset : nacl.dataset.TrainingDataset
        Data set of photometric and spectroscopic observations.
    """
    def __init__(self, filespath='../data/SALT3_training_sample/Data_reshape/',
                 filterpath='../data/SALT3_training_sample/filters'):
        """
        Constructor - computes the arguments

        Parameters
        ----------
        filespath : string
           path to the repository containing the light curve et spectra files.
        filterpath : string
           path to the filters files repository in nacl.dev/data/
        """
        self.filespath = filespath
        self.filterpath = filterpath

        self.Sne, self.snInfo, self.lc, self.sp = None, None, None, None
        self.sigma_lc, self.sigma_sp, self.trainingDataset = None, None, None

        self.data_type = [('Date', '<f8'), ('Flux', '<f8'), ('FluxErr', '<f8'),
                          ('Filter', '|S20'), ('Wavelength', '<f8'), 
                          ('MagSys', '|S20'), ('ZHelio', '<f8'), ('sn_id', '<i4'), ('spec_id', '<i4')]
        self.snInfo_type = [('z', '<f8'), ('tmax', '<f8'), ('x1', '<f8'),
                            ('x0', '<f8'), ('c', '<f8')]
        self.make_data()
        
    def merge_data_file(self):
        r"""
        Merge all the K21 survey training files of light curves and spectra.

        Returns
        -------
        lc_allsurveys : numpy.rec.array
             All surveys light curve file.
        lc_sum_allsurveys : numpy.rec.array
             All surveys light curve summary file.
        sp_allsurveys : numpy.rec.array
             All surveys spectra file.
        sp_sum_allsurveys : numpy.rec.array
             All surveys spectra summary file.
        """
        surveys = glob.glob(self.filespath + os.sep + '*_lc.npy')
        surveys = [i.split('/')[-1].split('_')[0] for i in surveys]
        lc_allsurveys, sp_allsurveys = [], []
        lc_sum_allsurveys, sp_sum_allsurveys = [], []
        
        for sur in surveys:
            lc = np.load(self.filespath + os.sep + f'{sur}_lc.npy')
            lc_sum = np.load(self.filespath + os.sep + f'{sur}_lc_summary.npy',
                             allow_pickle=True)
            try:
                sp = np.load(self.filespath + os.sep + f'{sur}_spectra.npy')
                sp_sum = np.load(self.filespath + os.sep + f'{sur}_spectra_summary.npy',
                                 allow_pickle=True)
            except FileNotFoundError:
                print('No spectra')
                sp = None
            try:
                lc_id_offset = lc_allsurveys[-1]['id']+1
                sp_id_offset = sp_allsurveys[-1]['id']+1
            except IndexError:
                lc_id_offset, sp_id_offset = 0, 0
   
            lc['id'] += lc_id_offset
            
            if sp is not None:
                sp['id'] += sp_id_offset
                
            if np.where(np.array(surveys) == sur)[0][0] == 0:
                lc_allsurveys = lc
                lc_sum_allsurveys = lc_sum
                if sp is not None:
                    sp_allsurveys = sp
                    sp_sum_allsurveys = sp_sum
                    
            else:
                lc_allsurveys = np.hstack((lc_allsurveys, lc))
                lc_sum_allsurveys = np.hstack((lc_sum_allsurveys, lc_sum))
                if sp is not None:
                    sp_allsurveys = np.hstack((sp_allsurveys, sp))
                    sp_sum_allsurveys = np.hstack((sp_sum_allsurveys, sp_sum))
                    
        return lc_allsurveys, lc_sum_allsurveys, sp_allsurveys, sp_sum_allsurveys

    @staticmethod
    def get_sn_parameter(lc_sum):
        r"""
        Get sn parameters from published files

        Parameters
        ----------
        lc_sum : array
            Light curve summary file.

        Returns
        -------
        x0 : list
            :math:`X_0` of K21 training sets
        x1 : list
            :math:`X_1` of K21 training sets
        color : list
            :math:`c` of K21 training sets
        """
        salt3 = np.recfromtxt('../data/SALT3_training_sample/SALT3TRAIN_K21_PUBLIC/SALT3_PARS_INIT.LIST')
        x0, x1, color = [], [], []
        for isn in lc_sum['SN']:
            idx = salt3['f0'] == isn[1:]       
            x0.append(salt3[idx]['f2'][0])
            x1.append(salt3[idx]['f3'][0])
            color.append(salt3[idx]['f4'][0])
        return x0, x1, color
    
    def make_data(self):
        r"""
        Make training set from jla*.npy files, where all JLA data are gathered.
        Calculated ratio of error measured on measurements.
        
        Remove data out from definition ranges.
        """

        lc, lc_sum, sp, sp_sum = self.merge_data_file()
        
        lc_sum = lc_sum[np.unique(lc['id'])]
        sp_sum = sp_sum[np.unique(sp['id'])]    
        self.Sne = lc_sum['SN'][np.unique(lc['id'])]    
        lc['id'] = reindex(lc['id'])
        sp['id'] = reindex(sp['id'])

        # self.sne = lc_sum['SN']

        lc_mean_wavelength = np.zeros_like(lc['id'])
        bands = [i.decode('UTF-8') for i in np.unique(lc['Filter'])]
        filterset = nacl.instruments.load_instrument(bands[0].split('::')[0],
                                                     path=self.filterpath)  # salt2.load_filters(bands)
        for bd in bands:
            idxmw = lc['Filter'] == bd.encode('UTF-8')
            band_k21 = bd.split('::')[1]
            lc_mean_wavelength[idxmw] = filterset.get_efficiency(band_k21).mean()

        self.lc = np.rec.fromarrays((lc['Date'], lc['Flux'], lc['FluxErr'],
                                     lc['Filter'], lc_mean_wavelength, 
                                     lc['MagSys'], lc_sum['ZHelio'][lc['id']], lc['id'],
                                     # len(np.unique(sp['id'])) + np.arange(len(lc['id']))
                                     -1 * np.ones_like(lc['Flux'])),
                                    dtype=self.data_type)
        sp_none = np.array([b''] * len(sp))
        sp_id = np.array([np.where(lc_sum['SN'] == i)[0][0] for i in sp_sum['SN']])[sp['id']]
                
        self.sp = np.rec.fromarrays((sp_sum['Date'][sp['id']], sp['Flux'],
                                     sp['FluxErr'], sp_none, sp['Wavelength'],
                                     sp_none, lc_sum['ZHelio'][sp_id], sp_id, sp['id']),
                                    dtype=self.data_type)

        self.sigma_lc = np.abs(self.lc['FluxErr']/self.lc['Flux'])
        self.sigma_sp = np.abs(self.sp['FluxErr']/self.sp['Flux'])
        x0, x1, color = self.get_sn_parameter(lc_sum)
        self.snInfo = np.rec.fromarrays((lc_sum['ZHelio'], lc_sum['DayMax'], x1, x0, color),
                                        dtype=self.snInfo_type)
        self.indexing()
        self.trainingDataset = TrainingDataset(self.lc, self.sp, self.snInfo,
                                               filterpath=self.filterpath)

    def indexing(self):
        """
        Reindex spectral ('spec_id', 'sn_id') and photometric data ('band_id', 'lc_id', 'sn_id'),
        when data have been removed.
        """
        n_lc, n_sp = len(self.lc), len(self.sp)
        # band indexation 
        dict_bd = {}
        _, idx_bd = np.unique(self.lc['Filter'], return_index=True)
        for i_bd, bd in enumerate(self.lc['Filter'][np.sort(idx_bd)]):
            dict_bd[bd] = i_bd
        id_bd = np.array([dict_bd[bd] for bd in self.lc['Filter']])

        # light curve indexation
        c = 0
        id_lc = np.ones(len(self.lc['Flux']))  # .astype(int)
        
        for i in range(self.lc['sn_id'][-1]+1):
            idx_sn = self.lc['sn_id'] == i
            lcs = self.lc[idx_sn]
            _, idx = np.unique(lcs["Filter"], return_index=True)
            for bd_sn in lcs['Filter'][np.sort(idx)]:
                id_lc[(self.lc['sn_id'] == i) & (self.lc['Filter'] == bd_sn)] = c  # [c]*len(lc[lc]))
                c += 1

        id_lc = np.hstack(np.array(id_lc))
        self.lc = np.lib.recfunctions.rec_append_fields(self.lc,
                                                        names=['lc_id', 'band_id', 'i'],
                                                        data=[id_lc.astype(int), id_bd, np.arange(n_lc)])
        i_sp = n_lc + np.arange(n_sp)
        sp_ones = np.ones_like(i_sp)
        
        self.sp = np.lib.recfunctions.rec_append_fields(self.sp,
                                                        names=['lc_id', 'band_id', 'i'],
                                                        data=[-1*sp_ones, -1*sp_ones,
                                                              n_lc + np.arange(n_sp)])



import numpy as np
from scipy.sparse import coo_matrix


class Constraints2D:
    r"""
    Local (sparse) constraints to apply to the spline part of the salt model.
    The SED model is defined as :

     .. math::
        S(\lambda, t) = X_0 [M_0(\lambda, p) + X_1 M_1(\lambda, p)] \, e^{CL(\lambda) \, C}

    where the 2D surfaces are defined on the same spline basis :

    .. math::
        M_{0|1}(\lambda, \mathrm{p}) = \sum_{k\ell} \theta_{k\ell}^{0|1} \mathcal{B}^t_k(\mathrm{p})
        \mathcal{B}^m_l(\mathrm{\lambda})

    Constraints are:

    .. math::
        \int M_0(\lambda, p = 0, ) T_B(\lambda) \frac{\lambda}{hc} d\lambda = 1 \\

         \int M_1(\lambda, p = 0, ) T_B(\lambda) \frac{\lambda}{hc} d\lambda = 1 \\

        \int \left.\frac{\partial M_0(\lambda, p = 0)}{\partial t}\right|_{t = t_{max}^{B^\star}}
        T_B(\lambda) \frac{\lambda}{hc} d\lambda = 0 \\

         \int \left.\frac{\partial M_1(\lambda, p = 0)}{\partial t}\right|_{t = t_{max}^{B^\star}}
         T_B(\lambda) \frac{\lambda}{hc} d\lambda = 0 \\

         \left<c\right> = 0 \\

         \left<X_1\right> = 0 \\

         \left<(X_1 - \left<X_1\right>)^2\right> = 1 \\

    Six first constraints are linear. They are implemented in the same form :

    .. math::
        C(\beta) =  \alpha - H_{pen} \beta

    For practical reasons, :math:`H_{pen}` and :math:`\alpha` are divided by the square root of the constraints
    quadratic penalty, :math:`\sigma_{pen} = \sqrt{\mu_{pen}}`

    The last one is a quadratic constraint then it is called : :math:`C^1(X_1)`
    Only the constraint involving the free parameters of the model are considered.

    Attributes
    ----------
    model : nacl.salt
        Model.
    pars : nacl.lib.fitparameters
        Model parameters.
    basis : nacl.lib.bspline
        Model basis 2D surfaces splines 2D basis (wavelength, phase)

    parameters_cons : list
        Free parameters of the model, parameters to constraints

    x1_var_real : bool
        If False, consider that the mean of the constraint on the :math:`X_1` is already true, which mean that

         .. math::
             C^1(X_1) = \left<(X_1)^2\right> = 1

        If True :

         .. math::
             C^1(X_1) = \left<(X_1 - \left<X_1\right>)^2\right> = 1

    vals : array
         Values of the :math:`\alpha` vector, divided by :math:`\sigma_{pen}`

    sig_vals : array
        Values of the :math:`\sigma_{pen}` vector.

    mu_pq : float
        Values of the linear constraint quadratic penalty amplitude.
    mu_pq_var_x1 :
        Values of the :math:`X_1` variance constraint quadratic penalty amplitude.

    H : scipy.sparse.coo_matrix
        :math:`H_{pen}` divided by the square root of the constraints quadratic penalty, :math:`\sigma_{pen}`
    """

    def __init__(self, model, mu_pq, parameters_cons=['M0', 'X0', 'tmax'],
                 mu_pq_var_x1=None, x1_var_real=False):
        r"""
        Constructor.
        Creation of the sparse matrix :math:`H_{pen}`.
        Since the constraints are linear, each line correspond to a constraint and the colon to the model parameters
        (: :math:`\beta`) involved in this constraint.

        Parameters
        ----------
        model : nacl.salt
            Model.
        mu_pq : float
            Values of the linear constraint quadratic penalty amplitude.
        parameters_cons : list
            Free parameters of the model, parameters to constraints
        mu_pq_var_x1 :
            Values of the :math:`X_1` variance constraint quadratic penalty amplitude.
        x1_var_real : bool
            If False, consider that the mean of the constraint on the :math:`X_1` is already true, which mean that

             .. math::
                 C^1(X_1) = \left< (X_1)^2 \right> = 1

            If True :

             .. math::
                 C^1(X_1) = \left< (X_1 - \left< X_1 \right> )^2 \right> = 1
        """
        self.model = model
        self.parameters_cons = parameters_cons
        self.x1_var_real = x1_var_real
        
        self.pars = self.model.pars
        self.basis = self.model.basis

        n = len(self.pars.full)
        pp = self.pars.copy()
        pp.release()
        nsn = len(pp['X0'].full)
        self.vals = []  # vals
        tr = self.model.filter_db.transmission_db['SWOPE::B']
        tr_eval, _ = self.model.filter_db.insert(tr, z=0)
        filter_z = self.model.gram.dot(tr_eval)
        comp_t = 0
        self.mu_pq = mu_pq
        if mu_pq_var_x1 is None:
            self.mu_pq_var_x1 = self.mu_pq
        else:
            self.mu_pq_var_x1 = mu_pq_var_x1
        sigma = 1/np.sqrt(self.mu_pq)
        self.sig_vals = []
        
        i = np.array([])
        j = np.array([])
        v = np.array([])
        
        if ('M0' in parameters_cons) & ('X0' in parameters_cons):
            # first constraint: S(phase = 0, band = B) == 1.
            surface0_eval = self.model.basis.by.eval(np.array([0.])).toarray()
            jacobian = coo_matrix(np.outer(surface0_eval, filter_z).ravel())
            i = np.hstack((i, pp['M0'].indexof(jacobian.col)))
            j = np.hstack((j, np.full(len(jacobian.col), comp_t)))
            v = np.hstack((v, jacobian.data.copy()))
            comp_t += 1
            self.vals.append(1)
            self.sig_vals.append(sigma)
            
        if ('M0' in parameters_cons) & ('tmax' in parameters_cons):
            # second constraint: S'(phase = 0) == 0. band = B
            surface_zero_derivative = self.model.basis.by.deriv(np.array([0.])).toarray()
            integral_zero_derivative = coo_matrix(np.outer(surface_zero_derivative, filter_z).ravel())
            i = np.hstack((i, pp['M0'].indexof(integral_zero_derivative.col)))
            j = np.hstack((j, np.full(len(integral_zero_derivative.col), comp_t)))
            v = np.hstack((v, integral_zero_derivative.data.copy()))
            comp_t += 1
            self.vals.append(0)
            self.sig_vals.append(sigma)
            
        if ('M1' in parameters_cons) & ('X1' in parameters_cons):
            # third constraint: flux_broadband_max  = 0
            surface1_eval = self.model.basis.by.eval(np.array([0.])).toarray()
            jacobian = coo_matrix(np.outer(surface1_eval, filter_z).ravel())
            i = np.hstack((i, pp['M1'].indexof(jacobian.col)))
            j = np.hstack((j, np.full(len(jacobian.col), comp_t)))
            v = np.hstack((v, jacobian.data.copy()))
            comp_t += 1
            self.vals.append(0)
            self.sig_vals.append(sigma)
                       
        if ('M1' in parameters_cons) & ('tmax' in parameters_cons):
            # fourth constraint: flux_broadband_max = 0
            surface1_zero_derivative = self.model.basis.by.deriv(np.array([0.])).toarray()
            integral1_zero_derivative = coo_matrix(np.outer(surface1_zero_derivative, filter_z).ravel())
            i = np.hstack((i, pp['M1'].indexof(integral1_zero_derivative.col)))
            j = np.hstack((j, np.full(len(integral1_zero_derivative.col), comp_t)))
            v = np.hstack((v, integral1_zero_derivative.data.copy()))
            comp_t += 1
            self.vals.append(0)
            self.sig_vals.append(sigma)
           
        v *= self.model.norm
        
        if 'X1' in parameters_cons:            
            # mean X1 = 0
            i = np.hstack((i, pp['X1'].indexof(np.arange(nsn))))
            j = np.hstack((j, np.full(nsn, comp_t)))
            v = np.hstack((v, np.full(nsn, 1)))  # X1 new constraint 1e-5
            comp_t += 1
            self.vals.append(0)
            self.sig_vals.append(sigma)
           
        if 'c' in parameters_cons:
            # mean c = 0
            i = np.hstack((i, pp['c'].indexof(np.arange(nsn))))
            j = np.hstack((j, np.full(nsn, comp_t)))
            v = np.hstack((v, np.full(nsn, 1)))
            comp_t += 1
            self.vals.append(0)
            self.sig_vals.append(sigma)

        self.cl_i = None
        self.sig_vals = np.array(self.sig_vals)
        self.vals = np.array(self.vals)/self.sig_vals
        v /= self.sig_vals[np.array(j).astype(int)]
        
        self.H = coo_matrix((v, (i, j)), shape=(n, len(self.vals))).tocsr()

        if 'X1' in self.parameters_cons:
            self.var_x1 = self.var_x1
        else:
            self.var_x1 = None
            
    def var_x1(self, beta, gamma=[], jac=False):
        r"""
        Evaluate the seventh constraints, on the variance of the :math:`X_1` parameters.

        If the attributes x1_var_real is True, then compute :
            the constraint :

            .. math::
                 C^1(X_1) = \left<(X_1 - \left<X_1\right>)^2\right> = 1

            it firsts derivatives wrt :math:`X_1` :

            .. math::
                \frac{dC(X_1)}{dX_1^i} = \frac{2}{N} (X_1 - \bar{X_1})^T \left(\begin{matrix}
                - \frac{1}{N} \\
                ... \\
                (1 - \frac{1}{N})_i \\
                ... \\
                - \frac{1}{N}
                \end{matrix} \right)

            it seconds derivatives wrt :math:`X_1` :

                .. math::
                    \frac{d^2C(X_1)}{dX_1^j dX_1^i} &= \frac{2}{N} \left(\begin{matrix}
                    - \frac{1}{N} \\
                    ... \\
                    (1 - \frac{1}{N})_j \\
                    ... \\
                    - \frac{1}{N}
                    \end{matrix} \right)^T  \left(\begin{matrix}
                    - \frac{1}{N} \\
                    ... \\
                    (1 - \frac{1}{N})_i \\
                    ... \\
                    - \frac{1}{N}
                    \end{matrix} \right)

        If the attributes x1_var_real is False :

        .. math::
            C^1(X_1) =  \frac{1}{N} X_1^T X_1 \quad  \& \quad
            \frac{\partial C^1(X_1)}{\partial X_1^i} = \frac{2}{N^2} X_1^T  \quad  \& \quad
            \frac{\partial^2 C^1(X_1)}{ \partial X_1^j \partial X_1^i} = \frac{2}{N^2} \delta{ij}


        Parameters
        ----------
        beta : array
            Model parameters.
        gamma : array
            Variance model parameters.
        jac : bool
            If derivatives are needed.

        Returns
        -------
        cons : array
            Value of the constraints multiply by the square root of mu_pq_var_x1

        if jac is True :
        jacobian_cons : array
            Gradient of the constraints multiply by mu_pq_var_x1
        hessian_cons :
            Hessian of the constraints multiply by mu_pq_var_x1
        """

        self.pars.free = beta
        pp = self.pars.copy()
        nsn = len(pp['X1'].full)
        diag = np.zeros_like(pp.free)
        i = pp['X1'].indexof(np.arange(nsn))
        xh, yh = np.meshgrid(i, i)
        if self.x1_var_real is False:
            diag[i] = np.ones(nsn)
            cons = pp['X1'].free.var() - 1
            dc0 = 2/nsn * diag * pp.free
            dc = np.matrix(dc0[i])
            jacobian_cons = 2 * dc0 * cons
            # if gamma is not None:
            # jacobian_cons = self.mu_pq * np.hstack((jacobian_cons, np.zeros(len(gamma))))
            jacobian_cons = self.mu_pq_var_x1 * np.hstack((jacobian_cons, np.zeros(len(gamma))))
            hh = np.diag([2/nsn] * len(i))
            # hessian_cons = self.mu_pq * (2 * np.matrix(hh) * cons + 2 * dc.T @ dc)
            hessian_cons = self.mu_pq_var_x1 * (2 * np.matrix(hh) * cons + 2 * dc.T @ dc)

        # ## new implementation
        # ## here mean is not Zero, mush longer for the fit to
        # ## reach the constraints but when we start far from
        # ## <X1> != 0 needed
        elif self.x1_var_real:
            jacobian_cons = np.zeros_like(pp.free)
            x1mean = pp['X1'].free.mean()
            x1v = pp['X1'].free - x1mean
            cons = pp['X1'].free.var() - 1  # np.mean(x1v**2) - 1

            dc_dxi = 2 / nsn ** 2 * (nsn * pp['X1'].free - (2 * nsn - 1) * x1mean)
            jacobian_cons[i] = 2 * cons * dc_dxi
            jacobian_cons = self.mu_pq_var_x1 * np.hstack((jacobian_cons, np.zeros(len(gamma))))

            dc_dxi_2 = 2 / nsn ** 2 * (nsn * np.diag(np.ones_like(x1v)) - (2 * nsn - 1) / nsn)
            hessian_cons = self.mu_pq_var_x1 * 2 * (dc_dxi.reshape(-1, 1) @ dc_dxi.reshape(-1, 1).T + cons * dc_dxi_2)

        idx = (yh.ravel() != -1) & (xh.ravel() != -1)
        hessian_cons = coo_matrix((np.array(hessian_cons).ravel()[idx],
                                   (yh.ravel()[idx], xh.ravel()[idx])),
                                  shape=(len(pp.free)+len(gamma), len(pp.free)+len(gamma)))
        cons *= np.sqrt(self.mu_pq_var_x1)
        if jac:
            return cons, jacobian_cons, hessian_cons
        return cons

    def get_rhs(self):
        i = np.where(self.pars.indexof() >= 0)[0]
        h = self.H[i, :].tocoo()
        idx = np.bincount(h.col, minlength=len(self.vals)) > 0
        return self.vals[idx]

    def __call__(self, beta, jac=False):
        r"""
        Evaluate the linear constraints: :math:`C(\beta) = \alpha - H_{pen} \beta`

        Parameters
        ----------
        beta : array
            Model parameters.
        jac : bool
            If derivatives are needed.

        Returns
        -------
        v : array
            Value of the constraints multiply by the square root of :math:`\mu_{pq}`
        if jac is True :
        h : array
            Matrix :math:`H_{pen}` by :math:`\mu_{pq}`
        """
        self.pars.free = beta            
        v = -self.H.T @ self.pars.full + self.vals
        # we need to evaluate how many constraints are still valid,
        # given which parameters are fixed. We do that by slicing the
        # constraint matrix. There is probably a better and faster
        # way, which avoids having to do this each time the
        # constraints are evaluated
        i = np.where(self.pars.indexof() >= 0)[0]
        h = self.H[i, :].tocoo()
        # verify that all columns of H are populated
        j = np.where(np.bincount(h.col) > 0)[0]
        h = h.tocsr()[:, j].tocoo()
        # we need also to slice the return values
        idx = np.bincount(h.col, minlength=len(self.vals)) > 0
        if jac == 1:
            return v[idx], h.T
        return v[idx]
